from __future__ import annotations

import importlib.resources
import json
import warnings
from typing import Any
from typing import Optional
from typing import cast

import jsonpath_ng.ext
import jsonschema
import numpy as np
import pandas as pd
import regex

from dsp_tools.commands.excel2json.models.input_error import ExcelFileProblem
from dsp_tools.commands.excel2json.models.input_error import InvalidExcelContentProblem
from dsp_tools.commands.excel2json.models.input_error import JsonValidationPropertyProblem
from dsp_tools.commands.excel2json.models.input_error import MissingValuesProblem
from dsp_tools.commands.excel2json.models.input_error import MoreThanOneSheetProblem
from dsp_tools.commands.excel2json.models.input_error import PositionInExcel
from dsp_tools.commands.excel2json.models.input_error import PropertyProblem
from dsp_tools.commands.excel2json.models.ontology import GuiAttributes
from dsp_tools.commands.excel2json.models.ontology import OntoProperty
from dsp_tools.commands.excel2json.utils import add_optional_columns
from dsp_tools.commands.excel2json.utils import check_column_for_duplicate
from dsp_tools.commands.excel2json.utils import check_contains_required_columns
from dsp_tools.commands.excel2json.utils import check_required_values
from dsp_tools.commands.excel2json.utils import col_must_or_not_empty_based_on_other_col
from dsp_tools.commands.excel2json.utils import find_one_full_cell_in_cols
from dsp_tools.commands.excel2json.utils import get_comments
from dsp_tools.commands.excel2json.utils import get_labels
from dsp_tools.commands.excel2json.utils import get_wrong_row_numbers
from dsp_tools.commands.excel2json.utils import read_and_clean_all_sheets
from dsp_tools.models.exceptions import InputError
from dsp_tools.models.problems import Problem

languages = ["en", "de", "fr", "it", "rm"]
language_label_col = ["label_en", "label_de", "label_fr", "label_it", "label_rm"]
mandatory_properties = ["name", "object", "gui_element"]


def excel2properties(
    excelfile: str,
    path_to_output_file: Optional[str] = None,
) -> tuple[list[dict[str, Any]], bool]:
    """
    Converts properties described in an Excel file into a "properties" section which can be inserted into a JSON
    project file.

    Args:
        excelfile: path to the Excel file containing the properties
        path_to_output_file: if provided, the output is written into this JSON file

    Raises:
        InputError: if something went wrong

    Returns:
        a tuple consisting of the "properties" section as a Python list,
            and the success status (True if everything went well)
    """

    property_df = _read_check_property_df(excelfile)

    property_df = _rename_deprecated_columnnames(df=property_df)

    # Not all columns have to be filled, users may delete some for ease of use, but it would generate an error later
    optional_col_set = {
        "label_en",
        "label_de",
        "label_fr",
        "label_it",
        "label_rm",
        "comment_en",
        "comment_de",
        "comment_fr",
        "comment_it",
        "comment_rm",
        "subject",
    }
    property_df = add_optional_columns(property_df, optional_col_set)

    _do_property_excel_compliance(df=property_df)

    _check_for_deprecated_syntax(property_df)

    # transform every row into a property
    props: list[OntoProperty] = []
    problems: list[Problem] = []
    for index, row in property_df.iterrows():
        result = _row2prop(
            df_row=row,
            row_num=int(str(index)) + 2,  # index is a label/index/hashable, but we need an int
        )
        if isinstance(result, PropertyProblem):
            problems.append(result)
        else:
            props.append(result)
    if problems:
        msg = ExcelFileProblem("properties.xlsx", problems).execute_error_protocol()
        raise InputError(msg)

    serialised_prop = [x.serialise() for x in props]

    # write final JSON file
    _validate_properties_section_in_json(properties_list=serialised_prop)
    if path_to_output_file:
        with open(file=path_to_output_file, mode="w", encoding="utf-8") as file:
            json.dump(serialised_prop, file, indent=4, ensure_ascii=False)
            print(f"properties section was created successfully and written to file '{path_to_output_file}'")

    return serialised_prop, True


def _check_for_deprecated_syntax(df: pd.DataFrame) -> None:
    pass


def _read_check_property_df(excelfile: str) -> pd.DataFrame:
    sheets_df_dict = read_and_clean_all_sheets(excelfile)
    if len(sheets_df_dict) != 1:
        msg = MoreThanOneSheetProblem("properties.xlsx", list(sheets_df_dict.keys())).execute_error_protocol()
        raise InputError(msg)
    return next(iter(sheets_df_dict.values()))


def _rename_deprecated_columnnames(df: pd.DataFrame) -> pd.DataFrame:
    df = _rename_deprecated_lang_cols(df=df)
    df = _rename_deprecated_hlist(df=df)
    return df


def _rename_deprecated_lang_cols(df: pd.DataFrame) -> pd.DataFrame:
    if set(language_label_col).issubset(set(df.columns)):
        return df
    if set(languages).issubset(set(df.columns)):
        warnings.warn(
            f"The file 'properties.xlsx' uses {languages} as column titles, which is deprecated. "
            f"Please use {[f'label_{lang}' for lang in languages]}"
        )
    rename_dict = dict(zip(languages, language_label_col))
    df = df.rename(columns=rename_dict)
    return df


def _rename_deprecated_hlist(df: pd.DataFrame) -> pd.DataFrame:
    # This function deals with Excel files that do conform to a previous format.
    if "hlist" not in df.columns:
        return df
    warnings.warn(
        "The file 'properties.xlsx' has a column 'hlist', which is deprecated. "
        "Please use the column 'gui_attributes' for the attribute 'hlist'."
    )
    df["hlist"] = df["hlist"].apply(lambda x: f"hlist:{x}" if isinstance(x, str) else x)
    # If gui_attributes already exists we have to merge the columns
    if "gui_attributes" in df.columns:
        # In case there is a hlist, it is the only valid value in gui_attributes and has precedence
        df["hlist"] = df["hlist"].fillna(df["gui_attributes"])
        df.pop("gui_attributes")
    df = df.rename(columns={"hlist": "gui_attributes"})
    return df


def _do_property_excel_compliance(df: pd.DataFrame) -> None:
    required_columns = {
        "name",
        "super",
        "object",
        "gui_element",
        "gui_attributes",
    }
    problems: list[Problem] = []
    if req_prob := check_contains_required_columns(df=df, required_columns=required_columns):
        problems.append(req_prob)
    if col_prob := check_column_for_duplicate(df=df, to_check_column="name"):
        problems.append(col_prob)
    if missing_vals_check := _check_missing_values_in_row(df=df):
        problems.append(missing_vals_check)
    if any(problems):
        excel_prob = ExcelFileProblem("properties.xlsx", problems)
        msg = excel_prob.execute_error_protocol()
        raise InputError(msg)


def _check_missing_values_in_row(df: pd.DataFrame) -> None | MissingValuesProblem:
    required_values = ["name", "super", "object", "gui_element"]
    missing_dict = check_required_values(df=df, required_values_columns=required_values)
    missing_labels = find_one_full_cell_in_cols(df=df, required_columns=language_label_col)
    if missing_labels is not None:
        missing_dict.update({"label": missing_labels})
    missing_gui_attributes = _check_compliance_gui_attributes(df=df)
    if missing_gui_attributes is not None:
        missing_dict.update(missing_gui_attributes)
    if missing_dict:
        missing_int_dict = get_wrong_row_numbers(wrong_row_dict=missing_dict, true_remains=True)
        locs = []
        for col, row_nums in missing_int_dict.items():
            locs.extend([PositionInExcel(column=col, row=x) for x in row_nums])
        return MissingValuesProblem(locs)
    else:
        return None


def _check_compliance_gui_attributes(df: pd.DataFrame) -> dict[str, pd.Series[bool]] | None:
    mandatory_attributes = ["Spinbox", "List"]
    mandatory_check = col_must_or_not_empty_based_on_other_col(
        df=df,
        substring_list=mandatory_attributes,
        substring_colname="gui_element",
        check_empty_colname="gui_attributes",
        must_have_value=True,
    )
    no_attributes = ["Checkbox", "Date", "Geonames", "Richtext", "TimeStamp"]
    no_attribute_check = col_must_or_not_empty_based_on_other_col(
        df=df,
        substring_list=no_attributes,
        substring_colname="gui_element",
        check_empty_colname="gui_attributes",
        must_have_value=False,
    )
    final_series = _get_final_series(mandatory_check, no_attribute_check)
    return {"gui_attributes": final_series} if final_series is not None else None


def _get_final_series(
    mandatory_check: pd.Series[bool] | None, no_attribute_check: pd.Series[bool] | None
) -> pd.Series[bool] | None:
    final_series: pd.Series[bool] = pd.Series()
    match mandatory_check, no_attribute_check:
        case None, None:
            return None
        case pd.Series(), pd.Series():
            final_series = pd.Series(np.logical_or(mandatory_check, no_attribute_check))
        case pd.Series(), None:
            final_series = mandatory_check
        case None, pd.Series():
            final_series = no_attribute_check
    return final_series


def _row2prop(df_row: pd.Series[Any], row_num: int) -> OntoProperty | PropertyProblem:
    subj = df_row["subject"] if not pd.isna(df_row["subject"]) else None
    comment = get_comments(df_row=df_row)
    gui_attrib = _get_gui_attribute(df_row=df_row, row_num=row_num)
    if isinstance(gui_attrib, InvalidExcelContentProblem):
        return PropertyProblem(df_row["name"], cast(list[Problem], [gui_attrib]))
    prop = OntoProperty(
        name=df_row["name"],
        super=[s.strip() for s in df_row["super"].split(",")],
        object=df_row["object"],
        subject=subj,
        labels=get_labels(df_row=df_row),
        comments=comment,
        gui_element=df_row["gui_element"],
        gui_attributes=gui_attrib,
    )
    return prop


def _get_gui_attribute(
    df_row: pd.Series[Any],
    row_num: int,
) -> GuiAttributes | InvalidExcelContentProblem | None:
    if pd.isnull(df_row["gui_attributes"]):
        return None
    # If the attribute is not in the correct format, a called function may raise an IndexError
    try:
        return _format_gui_attribute(attribute_str=df_row["gui_attributes"])
    except IndexError:
        return InvalidExcelContentProblem(
            expected_content="attribute: value, attribute: value",
            actual_content=df_row["gui_attributes"],
            excel_position=PositionInExcel(column="gui_attributes", row=row_num),
        )


def _format_gui_attribute(attribute_str: str) -> GuiAttributes:
    attribute_dict = _unpack_gui_attributes(attribute_str=attribute_str)
    return GuiAttributes(
        {attrib: _search_convert_numbers_in_str(value_str=val) for attrib, val in attribute_dict.items()}
    )


def _unpack_gui_attributes(attribute_str: str) -> dict[str, str]:
    gui_list = [x.strip() for x in attribute_str.split(",") if x.strip() != ""]
    sub_gui_list = [[sub.strip() for sub in x.split(":") if sub.strip() != ""] for x in gui_list]
    if any(len(sub) != 2 for sub in sub_gui_list):
        raise IndexError
    return {sub[0]: sub[1] for sub in sub_gui_list}


def _search_convert_numbers_in_str(value_str: str) -> str | int | float:
    if regex.search(r"^\d+$", value_str):
        return int(value_str)
    elif regex.search(r"^\d+\.\d+$", value_str):
        return float(value_str)
    else:
        return value_str


def _validate_properties_section_in_json(
    properties_list: list[dict[str, Any]],
) -> None:
    with (
        importlib.resources.files("dsp_tools")
        .joinpath("resources/schema/properties-only.json")
        .open(encoding="utf-8") as schema_file
    ):
        properties_schema = json.load(schema_file)
    try:
        jsonschema.validate(instance=properties_list, schema=properties_schema)
    except jsonschema.ValidationError as err:
        validation_problem = _find_validation_problem(properties_list=properties_list, validation_error=err)
        msg = ExcelFileProblem("properties.xlsx", [validation_problem]).execute_error_protocol()
        raise InputError(msg) from None


def _find_validation_problem(
    properties_list: list[dict[str, Any]],
    validation_error: jsonschema.ValidationError,
) -> JsonValidationPropertyProblem:
    if json_path_to_property := regex.search(r"^\$\[(\d+)\]", validation_error.json_path):
        # fmt: off
        wrong_property_name = (
            jsonpath_ng.ext.parse(json_path_to_property.group(0))
            .find(properties_list)[0]
            .value["name"]
        )
        # fmt: on
        excel_row = int(json_path_to_property.group(1)) + 2

        column = None
        val_msg = None
        if affected_field := regex.search(
            r"name|labels|comments|super|subject|object|gui_element|gui_attributes",
            validation_error.json_path,
        ):
            column = affected_field.group(0)
            val_msg = validation_error.message

        return JsonValidationPropertyProblem(
            problematic_property=wrong_property_name,
            excel_position=PositionInExcel(column=column, row=excel_row),
            original_msg=val_msg,
        )
    return JsonValidationPropertyProblem(
        original_msg=validation_error.message,
        message_path=validation_error.json_path,
    )



"""
The code in this file handles the arguments passed by the user from the command line and calls the requested actions.
"""

import argparse
import subprocess
import sys
from importlib.metadata import version

import regex
import requests
from loguru import logger
from packaging.version import parse

from dsp_tools.cli.call_action import call_requested_action
from dsp_tools.cli.create_parsers import make_parser
from dsp_tools.models.exceptions import BaseError
from dsp_tools.models.exceptions import InternalError
from dsp_tools.models.exceptions import UserError
from dsp_tools.utils.ansi_colors import BOLD_RED
from dsp_tools.utils.ansi_colors import RESET_TO_DEFAULT
from dsp_tools.utils.logger_config import logger_config
from dsp_tools.utils.warnings_config import initialize_warnings


def main() -> None:
    """
    Main entry point of the program as referenced in pyproject.toml
    """
    logger_config()
    run(sys.argv[1:])


def run(args: list[str]) -> None:
    """
    Main function of the CLI.

    Args:
        args: a list of arguments passed by the user from the command line,
            excluding the leading "dsp-tools" command.

    Raises:
        UserError: if user input was wrong
        InputError: if user input was wrong
        InternalError: if the user cannot fix it
    """
    default_dsp_api_url = "http://0.0.0.0:3333"
    default_dsp_ingest_url = "http://0.0.0.0:3340"
    root_user_email = "root@example.com"
    root_user_pw = "test"

    parser = make_parser(
        default_dsp_api_url=default_dsp_api_url,
        root_user_email=root_user_email,
        root_user_pw=root_user_pw,
    )
    parsed_arguments = _parse_arguments(
        user_args=args,
        parser=parser,
    )
    if not parsed_arguments.suppress_update_prompt:
        _check_version()
    _log_cli_arguments(parsed_arguments)
    initialize_warnings()

    try:
        parsed_arguments = _derive_dsp_ingest_url(
            parsed_arguments=parsed_arguments,
            default_dsp_api_url=default_dsp_api_url,
            default_dsp_ingest_url=default_dsp_ingest_url,
        )
        success = call_requested_action(parsed_arguments)
    except BaseError as err:
        logger.exception("The process was terminated because of an Error:")
        print("\nThe process was terminated because of an Error:")
        print(err.message)
        sys.exit(1)
    except Exception as err:  # noqa: BLE001 (blind-except)
        logger.exception(err)
        print(InternalError())
        sys.exit(1)

    if not success:
        logger.error("Terminate without success")
        sys.exit(1)


def _check_version() -> None:
    """
    Check if the installed version of dsp-tools is up-to-date.
    If pypi.org cannot be reached, or if the latest version cannot be determined, do nothing.
    If a new post-release version is available, print a message to the user, and return None.
    If the base version (i.e. the major.minor.micro part of the version) is outdated,
    ask the user if they want to exit or continue anyway.
    """
    try:
        response = requests.get("https://pypi.org/pypi/dsp-tools/json", timeout=5)
    except (requests.ConnectionError, requests.ReadTimeout):
        return
    if not response.ok:
        return
    latest = parse(response.json()["info"]["version"])
    installed = parse(version("dsp-tools"))
    if latest <= installed:  # in the release-please PR, the installed version is always greater than the latest
        return

    msg = (
        f"You are using DSP-TOOLS version {installed}, but version {latest} is available. "
        "Consider upgrading via 'pip3 install --upgrade dsp-tools'.\n"
    )
    if latest.base_version == installed.base_version:
        print(BOLD_RED + msg + RESET_TO_DEFAULT)
        return

    msg += "Continue anyway? [y/n]"
    resp = None
    while resp not in ["y", "n"]:
        resp = input(BOLD_RED + msg + RESET_TO_DEFAULT)
    if resp == "y":
        return
    sys.exit(1)


def _parse_arguments(
    user_args: list[str],
    parser: argparse.ArgumentParser,
) -> argparse.Namespace:
    """
    Parse the user-provided CLI arguments.
    If no action is provided,
    print the help text and exit with error code 1.

    Args:
        user_args: user-provided CLI arguments
        parser: parser used to parse the arguments

    Returns:
        parsed arguments
    """
    args = parser.parse_args(user_args)
    if not hasattr(args, "action"):
        parser.print_help(sys.stderr)
        sys.exit(1)
    return args


def _get_version() -> str:
    try:
        pip_output = subprocess.run("pip freeze".split(), check=False, capture_output=True).stdout.decode("utf-8")
        dsp_tools_lines = [x for x in pip_output.split("\n") if "dsp-tools" in x]
    except FileNotFoundError:
        dsp_tools_lines = []
    if not dsp_tools_lines:
        # if the virtual environment was activated with env variables instead of executing activation commands,
        # dsp-tools will run correctly, but "pip freeze" won't show dsp-tools
        return version("dsp-tools")
    _detail_version = dsp_tools_lines[0]
    # _detail_version has one of the following formats:
    # - 'dsp-tools==5.0.3\n'
    # - 'dsp-tools==5.6.0.post9\n'
    # - 'dsp-tools @ git+https://github.com/dasch-swiss/dsp-tools.git@1f95f8d1b79bd5170a652c0d04e7ada417d76734\n'
    # - '-e git+ssh://git@github.com/dasch-swiss/dsp-tools.git@af9a35692b542676f2aa0a802ca7fc3b35f5713d#egg=dsp_tools\n'
    # - ''
    if version_number := regex.search(r"\d+\.\d+\.\d+(\.(post|dev|pre)\d+)?", _detail_version):
        return version_number.group(0)
    if regex.search(r"github.com", _detail_version):
        return _detail_version.replace("\n", "")
    return version("dsp-tools")


def _log_cli_arguments(parsed_args: argparse.Namespace) -> None:
    """
    Log the CLI arguments passed by the user from the command line.

    Args:
        parsed_args: parsed arguments
    """
    metadata_lines = [
        f"DSP-TOOLS: Called the action '{parsed_args.action}' from the command line",
        f"DSP-TOOLS version: {_get_version()}",
        f"Location of this installation: {__file__}",
        "CLI arguments:",
    ]
    metadata_lines = [f"*** {line}" for line in metadata_lines]

    parameter_lines = []
    parameters_to_log = {key: value for key, value in vars(parsed_args).items() if key != "action"}
    longest_key_length = max((len(key) for key in parameters_to_log), default=0)
    for key, value in parameters_to_log.items():
        if key == "password":
            parameter_lines.append(f"{key:<{longest_key_length}} = {'*' * len(value)}")
        else:
            parameter_lines.append(f"{key:<{longest_key_length}} = {value}")
    parameter_lines = parameter_lines or ["(no parameters)"]
    parameter_lines = [f"***   {line}" for line in parameter_lines]

    asterisk_count = max(len(line) for line in metadata_lines + parameter_lines)
    logger.info("*" * asterisk_count)
    for line in metadata_lines:
        logger.info(line)
    for line in parameter_lines:
        logger.info(line)
    logger.info("*" * asterisk_count)


def _get_canonical_server_and_dsp_ingest_url(
    server: str,
    default_dsp_api_url: str,
    default_dsp_ingest_url: str,
) -> tuple[str, str]:
    """
    Based on the DSP server URL passed by the user,
    transform it to its canonical form,
    and derive the ingest server URL from it.

    If the DSP server URL points to port 3333 on localhost,
    the ingest server will point to port 3340 on localhost.

    If the DSP server URL points to a remote server ending in "dasch.swiss",
    modify it (if necessary) to point to the "api" subdomain of that server,
    and add a new "dsp_ingest_url" argument pointing to the "ingest" subdomain of that server.

    Args:
        server: DSP server URL passed by the user
        default_dsp_api_url: default DSP server on localhost
        default_dsp_ingest_url: default ingest server on localhost

    Raises:
        UserError: if the DSP server URL passed by the user is invalid

    Returns:
        canonical DSP URL and ingest server URL
    """
    localhost_match = regex.search(r"(0\.0\.0\.0|localhost):3333", server)
    remote_url_match = regex.search(
        r"^(?:https?:\/\/)?(?:admin\.|api\.|ingest\.|app\.)?((?:.+\.)?dasch)\.swiss", server
    )

    if localhost_match:
        server = default_dsp_api_url
        dsp_ingest_url = default_dsp_ingest_url
    elif remote_url_match:
        server = f"https://api.{remote_url_match.group(1)}.swiss"
        dsp_ingest_url = f"https://ingest.{remote_url_match.group(1)}.swiss"
    else:
        logger.error(f"Invalid DSP server URL '{server}'")
        raise UserError(f"ERROR: Invalid DSP server URL '{server}'")

    logger.info(f"Using DSP server '{server}' and ingest server '{dsp_ingest_url}'")
    print(f"Using DSP server '{server}' and ingest server '{dsp_ingest_url}'")

    return server, dsp_ingest_url


def _derive_dsp_ingest_url(
    parsed_arguments: argparse.Namespace,
    default_dsp_api_url: str,
    default_dsp_ingest_url: str,
) -> argparse.Namespace:
    """
    Modify the parsed arguments so that the DSP and ingest server URLs are correct.
    Based on the DSP server URL passed by the user,
    transform it to its canonical form,
    and derive the ingest server URL from it.

    Args:
        parsed_arguments: CLI arguments passed by the user, parsed by argparse
        default_dsp_api_url: default DSP server on localhost
        default_dsp_ingest_url: default ingest server on localhost

    Raises:
        UserError: if the DSP server URL passed by the user is invalid

    Returns:
        the modified arguments
    """
    if not hasattr(parsed_arguments, "server"):
        # some CLI actions (like excel2json, excel2xml, start-stack, ...) don't have a server at all
        return parsed_arguments

    server, dsp_ingest_url = _get_canonical_server_and_dsp_ingest_url(
        server=parsed_arguments.server,
        default_dsp_api_url=default_dsp_api_url,
        default_dsp_ingest_url=default_dsp_ingest_url,
    )
    parsed_arguments.server = server
    parsed_arguments.dsp_ingest_url = dsp_ingest_url

    return parsed_arguments


if __name__ == "__main__":
    main()



import warnings

import pytest

from dsp_tools.xmllib.models.dsp_base_resources import RegionResource


class TestRegion:
    def test_serialise_no_warnings(self) -> None:
        region = RegionResource.create_new("res_id", "label", "region_of")
        region = region.add_rectangle((0.1, 0.1), (0.2, 0.2))
        with warnings.catch_warnings(record=True) as caught_warnings:
            region.serialise()
            assert len(caught_warnings) == 0

    def test_serialise_no_region(self) -> None:
        region = RegionResource.create_new("res_id", "label", "region_of")
        with warnings.catch_warnings(record=True) as caught_warnings:
            region.serialise()
            assert len(caught_warnings) == 1


if __name__ == "__main__":
    pytest.main([__file__])



from dataclasses import dataclass
from dataclasses import field
from typing import Any
from uuid import uuid4

from dsp_tools.commands.xmlupload.iri_resolver import IriResolver
from dsp_tools.commands.xmlupload.models.formatted_text_value import FormattedTextValue
from dsp_tools.commands.xmlupload.models.upload_state import UploadState
from dsp_tools.commands.xmlupload.project_client import ProjectInfo
from dsp_tools.commands.xmlupload.stash.stash_models import LinkValueStash
from dsp_tools.commands.xmlupload.stash.stash_models import LinkValueStashItem
from dsp_tools.commands.xmlupload.stash.stash_models import StandoffStash
from dsp_tools.commands.xmlupload.stash.stash_models import StandoffStashItem
from dsp_tools.commands.xmlupload.stash.stash_models import Stash
from dsp_tools.commands.xmlupload.upload_config import UploadConfig
from dsp_tools.commands.xmlupload.xmlupload import _upload_stash
from dsp_tools.utils.connection import Connection
from test.integration.commands.xmlupload.connection_mock import ConnectionMockBase

# ruff: noqa: ARG002 (unused-method-argument)
# ruff: noqa: D102 (undocumented-public-method)


@dataclass
class ProjectClientStub:
    """Stub class for ProjectClient."""

    con: Connection
    shortcode: str
    project_info: ProjectInfo | None

    def get_project_iri(self) -> str:
        raise NotImplementedError("get_project_iri not implemented")

    def get_ontology_iris(self) -> list[str]:
        raise NotImplementedError("get_project_iri not implemented")

    def get_ontology_name_dict(self) -> dict[str, str]:
        return {}

    def get_ontology_iri_dict(self) -> dict[str, str]:
        raise NotImplementedError("get_project_iri not implemented")


@dataclass
class ConnectionMock(ConnectionMockBase):
    """Mock class for Connection."""

    get_responses: list[dict[str, Any]] = field(default_factory=list)
    post_responses: list[dict[str, Any]] = field(default_factory=list)
    put_responses: list[dict[str, Any]] = field(default_factory=list)

    def get(
        self,
        route: str,
        headers: dict[str, str] | None = None,
    ) -> dict[str, Any]:
        return self.get_responses.pop(0)

    def post(
        self,
        route: str,
        data: dict[str, Any] | None = None,
        files: dict[str, tuple[str, Any]] | None = None,
        headers: dict[str, str] | None = None,
        timeout: int | None = None,
    ) -> dict[str, Any]:
        return self.post_responses.pop(0)

    def put(
        self,
        route: str,
        data: dict[str, Any] | None = None,
        headers: dict[str, str] | None = None,
    ) -> dict[str, Any]:
        return self.put_responses.pop(0)


class TestUploadLinkValueStashes:
    def test_upload_link_value_stash(self) -> None:
        """Upload stashed link values (resptr), if all goes well."""
        stash = Stash.make(
            standoff_stash=None,
            link_value_stash=LinkValueStash.make(
                [
                    LinkValueStashItem("001", "sometype", "someprop", "002"),
                ],
            ),
        )
        assert stash
        iri_resolver = IriResolver(
            {
                "001": "http://www.rdfh.ch/0001/001",
                "002": "http://www.rdfh.ch/0001/002",
            }
        )
        con: Connection = ConnectionMock(post_responses=[{}])
        upload_state = UploadState([], stash, UploadConfig(), {}, [], iri_resolver)
        _upload_stash(upload_state, ProjectClientStub(con, "1234", None))
        assert not upload_state.pending_stash or upload_state.pending_stash.is_empty()

    def test_upload_link_value_stash_multiple(self) -> None:
        """Upload multiple stashed link values (resptr), if all goes well."""
        stash = Stash.make(
            standoff_stash=None,
            link_value_stash=LinkValueStash.make(
                [
                    LinkValueStashItem("001", "sometype", "someprop", "002"),
                    LinkValueStashItem("001", "sometype", "someprop", "003"),
                    LinkValueStashItem("002", "sometype", "someprop", "003"),
                    LinkValueStashItem("004", "sometype", "someprop", "002"),
                ],
            ),
        )
        assert stash
        iri_resolver = IriResolver(
            {
                "001": "http://www.rdfh.ch/0001/001",
                "002": "http://www.rdfh.ch/0001/002",
                "003": "http://www.rdfh.ch/0001/003",
                "004": "http://www.rdfh.ch/0001/004",
            }
        )
        con: Connection = ConnectionMock(post_responses=[{}, {}, {}, {}])
        upload_state = UploadState([], stash, UploadConfig(), {}, [], iri_resolver)
        _upload_stash(upload_state, ProjectClientStub(con, "1234", None))
        assert not upload_state.pending_stash or upload_state.pending_stash.is_empty()


class TestUploadTextValueStashes:
    def test_upload_text_value_stash(self) -> None:
        """Upload stashed text values (standoff), if all goes well."""
        value_uuid = str(uuid4())
        property_name = "someprop"
        stash = Stash.make(
            standoff_stash=StandoffStash.make(
                [
                    StandoffStashItem(
                        "001", "sometype", value_uuid, property_name, FormattedTextValue("<p>some text</p>")
                    ),
                ]
            ),
            link_value_stash=None,
        )
        assert stash
        iri_resolver = IriResolver(
            {
                "001": "http://www.rdfh.ch/0001/001",
                "002": "http://www.rdfh.ch/0001/002",
            }
        )
        con: Connection = ConnectionMock(
            get_responses=[
                {
                    property_name: [
                        {
                            "@id": "http://www.rdfh.ch/0001/001/values/01",
                            "knora-api:textValueAsXml": "<p>not relevant</p>",
                        },
                        {
                            "@id": "http://www.rdfh.ch/0001/001/values/01",
                            "knora-api:textValueAsXml": f"<p>{value_uuid}</p>",
                        },
                    ],
                    "@context": {},
                },
            ],
            put_responses=[{}],
        )
        upload_state = UploadState([], stash, UploadConfig(), {}, [], iri_resolver)
        _upload_stash(upload_state, ProjectClientStub(con, "1234", None))
        assert not upload_state.pending_stash or upload_state.pending_stash.is_empty()

    def test_not_upload_text_value_stash_if_uuid_not_on_value(self) -> None:
        """
        Do not upload stashed text values (standoff), if the resource has no value containing the UUID of the stashed
        text value in its text.
        """
        value_uuid = str(uuid4())
        property_name = "someprop"
        stash = Stash.make(
            standoff_stash=StandoffStash.make(
                [
                    StandoffStashItem(
                        "001", "sometype", value_uuid, property_name, FormattedTextValue("<p>some text</p>")
                    ),
                ]
            ),
            link_value_stash=None,
        )
        assert stash
        iri_resolver = IriResolver(
            {
                "001": "http://www.rdfh.ch/0001/001",
                "002": "http://www.rdfh.ch/0001/002",
            }
        )
        con: Connection = ConnectionMock(
            get_responses=[
                {
                    property_name: [
                        {
                            "@id": "http://www.rdfh.ch/0001/001/values/01",
                            "knora-api:textValueAsXml": "<p>not relevant</p>",
                        },
                    ],
                    "@context": {},
                },
            ],
            put_responses=[{}],
        )
        upload_state = UploadState([], stash, UploadConfig(), {}, [], iri_resolver)
        _upload_stash(upload_state, ProjectClientStub(con, "1234", None))
        assert upload_state.pending_stash == stash



import pytest

from dsp_tools.commands.xmlupload.models.deserialise.deserialise_value import IIIFUriInfo
from dsp_tools.commands.xmlupload.models.deserialise.deserialise_value import XMLBitstream
from dsp_tools.commands.xmlupload.models.deserialise.deserialise_value import XMLProperty
from dsp_tools.commands.xmlupload.models.deserialise.deserialise_value import XMLValue
from dsp_tools.commands.xmlupload.models.deserialise.xmlresource import XMLResource
from dsp_tools.commands.xmlupload.models.formatted_text_value import FormattedTextValue
from dsp_tools.commands.xmlupload.models.lookup_models import IntermediaryLookups
from dsp_tools.commands.xmlupload.models.permission import Permissions
from dsp_tools.commands.xmlupload.models.permission import PermissionValue
from dsp_tools.models.datetimestamp import DateTimeStamp


@pytest.fixture
def lookups() -> IntermediaryLookups:
    return IntermediaryLookups(
        permissions={"open": Permissions({PermissionValue.CR: ["knora-admin:ProjectAdmin"]})},
        listnodes={"list:node": "http://rdfh.ch/9999/node"},
        namespaces={
            "knora-api": "http://api.knora.org/ontology/knora-api/v2#",
            "onto": "http://0.0.0.0:3333/ontology/9999/onto/v2#",
        },
    )


@pytest.fixture
def bool_prop() -> XMLProperty:
    return XMLProperty(name="onto:boolProp", valtype="boolean", values=[XMLValue(value="true")])


@pytest.fixture
def bool_prop_with_permissions() -> XMLProperty:
    return XMLProperty(name="onto:boolProp", valtype="boolean", values=[XMLValue(value="true", permissions="open")])


@pytest.fixture
def bool_prop_with_non_existing_permissions() -> XMLProperty:
    return XMLProperty(
        name="onto:boolProp", valtype="boolean", values=[XMLValue(value="true", permissions="nonExisting")]
    )


@pytest.fixture
def bool_prop_with_comment() -> XMLProperty:
    return XMLProperty(name="onto:boolProp", valtype="boolean", values=[XMLValue(value="true", comment="comment")])


@pytest.fixture
def color_prop() -> XMLProperty:
    return XMLProperty(name="onto:colorProp", valtype="color", values=[XMLValue("#5d1f1e")])


@pytest.fixture
def date_prop() -> XMLProperty:
    return XMLProperty(name="onto:dateProp", valtype="date", values=[XMLValue("CE:1849:CE:1850")])


@pytest.fixture
def decimal_prop() -> XMLProperty:
    return XMLProperty(name="onto:decimalProp", valtype="decimal", values=[XMLValue("1.4")])


@pytest.fixture
def decimal_prop_with_two_values() -> XMLProperty:
    return XMLProperty(name="onto:decimalProp", valtype="decimal", values=[XMLValue("1.0"), XMLValue("2.0")])


@pytest.fixture
def simple_text_prop() -> XMLProperty:
    return XMLProperty(name="onto:simpleTextProp", valtype="text", values=[XMLValue(value="text")])


@pytest.fixture
def richtext_prop() -> XMLProperty:
    return XMLProperty(
        name="onto:richTextProp",
        valtype="text",
        values=[XMLValue(FormattedTextValue("<text>this is text</text>"), resrefs={"id"})],
    )


@pytest.fixture
def geoname_prop() -> XMLProperty:
    return XMLProperty(name="onto:geonameProp", valtype="geoname", values=[XMLValue("5416656")])


@pytest.fixture
def integer_prop() -> XMLProperty:
    return XMLProperty(name="onto:integerProp", valtype="integer", values=[XMLValue("1")])


@pytest.fixture
def list_prop() -> XMLProperty:
    return XMLProperty(name="onto:listProp", valtype="list", values=[XMLValue("list:node")])


@pytest.fixture
def resptr_prop() -> XMLProperty:
    return XMLProperty(name="onto:linkProp", valtype="resptr", values=[XMLValue("other_id")])


@pytest.fixture
def time_prop() -> XMLProperty:
    return XMLProperty(name="onto:timeProp", valtype="time", values=[XMLValue("2019-10-23T13:45:12.01-14:00")])


@pytest.fixture
def uri_prop() -> XMLProperty:
    return XMLProperty(name="onto:uriProp", valtype="uri", values=[XMLValue("https://dasch.swiss")])


@pytest.fixture
def iiif_uri() -> IIIFUriInfo:
    return IIIFUriInfo("https://this/is/a/uri.jpg")


@pytest.fixture
def bitstream() -> XMLBitstream:
    return XMLBitstream("file.jpg")


@pytest.fixture
def iiif_uri_with_permission() -> IIIFUriInfo:
    return IIIFUriInfo("https://this/is/a/uri.jpg", "open")


@pytest.fixture
def bitstream_with_permission() -> XMLBitstream:
    return XMLBitstream("file.jpg", "open")


@pytest.fixture
def resource_one_prop(bool_prop: XMLProperty) -> XMLResource:
    return XMLResource(
        res_id="id",
        iri=None,
        ark=None,
        label="lbl",
        restype="onto:ResourceType",
        permissions=None,
        creation_date=None,
        bitstream=None,
        iiif_uri=None,
        properties=[bool_prop],
    )


@pytest.fixture
def resource_with_permissions() -> XMLResource:
    return XMLResource(
        res_id="id",
        iri=None,
        ark=None,
        label="lbl",
        restype="onto:ResourceType",
        permissions="open",
        creation_date=None,
        bitstream=None,
        iiif_uri=None,
        properties=[],
    )


@pytest.fixture
def resource_with_iiif_uri(iiif_uri: IIIFUriInfo) -> XMLResource:
    return XMLResource(
        res_id="id",
        iri=None,
        ark=None,
        label="lbl",
        restype="onto:ResourceType",
        permissions=None,
        creation_date=None,
        bitstream=None,
        iiif_uri=iiif_uri,
        properties=[],
    )


@pytest.fixture
def resource_with_bitstream(bitstream: XMLBitstream) -> XMLResource:
    return XMLResource(
        res_id="id",
        iri=None,
        ark=None,
        label="lbl",
        restype="onto:ResourceType",
        permissions=None,
        creation_date=None,
        bitstream=bitstream,
        iiif_uri=None,
        properties=[],
    )


@pytest.fixture
def resource_with_unknown_permissions() -> XMLResource:
    return XMLResource(
        res_id="id",
        iri=None,
        ark=None,
        label="lbl",
        restype="onto:ResourceType",
        permissions="nonExisting",
        creation_date=None,
        bitstream=None,
        iiif_uri=None,
        properties=[],
    )


@pytest.fixture
def resource_with_ark() -> XMLResource:
    return XMLResource(
        res_id="id",
        iri=None,
        ark="ark:/72163/4123-43xc6ivb931-a.2022829",
        label="lbl",
        restype="onto:ResourceType",
        permissions=None,
        creation_date=DateTimeStamp("1999-12-31T23:59:59.9999999+01:00"),
        bitstream=None,
        iiif_uri=None,
        properties=[],
    )


@pytest.fixture
def resource_with_iri() -> XMLResource:
    return XMLResource(
        res_id="id",
        iri="http://rdfh.ch/4123/DiAmYQzQSzC7cdTo6OJMYA",
        ark=None,
        label="lbl",
        restype="onto:ResourceType",
        permissions=None,
        creation_date=None,
        bitstream=None,
        iiif_uri=None,
        properties=[],
    )


@pytest.fixture
def resource_with_ark_and_iri() -> XMLResource:
    return XMLResource(
        res_id="id",
        iri="http://rdfh.ch/4123/DiAmYQzQSzC7cdTo6OJMYA",
        ark="ark:/72163/4123-43xc6ivb931-a.2022829",
        label="lbl",
        restype="onto:ResourceType",
        permissions=None,
        creation_date=DateTimeStamp("1999-12-31T23:59:59.9999999+01:00"),
        bitstream=None,
        iiif_uri=None,
        properties=[],
    )



from pathlib import Path
from typing import Any

from lxml import etree

from dsp_tools import excel2xml


def create_and_save_circular_references_test_graph(replication_counter: int = 1, save_location: str = "") -> None:
    """
    This function creates a graph with circular references.
    By default, it saves it in the current directory; this can be changed with the parameter save_location.

    Args:
        replication_counter: number of times the sub-graphs should be created in one root-graph
        save_location: path to the folder, where the file should be saved
    """
    root = excel2xml.make_root("0700", "simcir")
    root = create_circular_references_test_graph(root, replication_counter)
    excel2xml.write_xml(root, Path(save_location) / f"test_circular_references_{replication_counter}.xml")


def create_circular_references_test_graph(root: etree._Element, replication_counter: int) -> etree._Element:
    """
    This function creates a graph with circular references.
    It is capable of reproducing one graph with all the references a specified number of times.
    Each circle type has its own sub-graph number in order to differentiate them by their ID.
    The resulting etree is suitable for an upload to the DSP-API.

    Args:
        root: root of the XML
        replication_counter: number of times the sub-graphs should be created in one root-graph.

    Returns:
        An etree which is suitable for an upload into the DSP-API
    """
    for i in range(1, replication_counter + 1):
        root.extend(_make_one_circle_with_three_resources(replication_counter=f"{i}1"))
        root.extend(_make_complex_dependencies_single_link(replication_counter=f"{i}2.1"))
        root.extend(_make_complex_circular_dependencies_multi_link(replication_counter=f"{i}2.2"))
        root.extend(_make_complex_dependencies_add_on(replication_counter=f"{i}3"))
        root.extend(_make_two_references(replication_counter=f"{i}4.1"))
        root.extend(_make_multi_link_two_resource_circle(replication_counter=f"{i}4.2"))
        root.extend(_make_chain(replication_counter=f"{i}5"))
        root.extend(_make_inverted_complex_dependencies(replication_counter=f"{i}6"))
        root.extend(_make_two_resource_circle_plus_non_circle_link(replication_counter=f"{i}7"))
        root.extend(_make_three_resource_circle_with_multiple_text_prop(replication_counter=f"{i}8"))
        root.extend(_make_three_resource_circle_multiple_diverse_links(replication_counter=f"{i}9"))
        root.extend(_make_complex_circle_with_leaf_nodes(replication_counter=f"{i}10"))
    return root


def _make_list_of_resources(number_of_resources: int, replication_counter: str, start_range: int = 65) -> list[Any]:
    letters = [chr(x) for x in range(start_range, start_range + number_of_resources)]
    return [
        excel2xml.make_resource(
            restype=":TestThing", label=f"res_{let}_{replication_counter}", id=f"res_{let}_{replication_counter}"
        )
        for let in letters
    ]


def _make_salsah_link(target_id: str) -> str:
    return f'<a class="salsah-link" href="IRI:{target_id}:IRI">{target_id}</a>'


def _make_xml_text_prop(target_res: etree._Element | list[etree._Element]) -> etree._Element:
    match target_res:
        case etree._Element():
            salsah_link = _make_salsah_link(target_res.attrib["id"])
            # one resource with many targets
        case list():
            salsah_link = "blabla".join([_make_salsah_link(x.attrib["id"]) for x in target_res])
    text_value = f"Start text{salsah_link}end text."  # type: ignore[possibly-undefined]
    return excel2xml.make_text_prop(name=":hasRichtext", value=excel2xml.PropertyElement(text_value, encoding="xml"))


def _make_resptr_prop(target_id: list[str] | str, property_name: str = ":hasResource1") -> etree._Element:
    match target_id:
        case str():
            link = excel2xml.make_resptr_prop(name=property_name, value=target_id)
        # one resource with many targets
        case list():
            link = excel2xml.make_resptr_prop(name=property_name, value=target_id)
    return link  # type: ignore[possibly-undefined]


def _make_simple_text(resource: etree._Element) -> etree._Element:
    resource.append(excel2xml.make_text_prop(name=":hasSimpleText", value="This is text."))
    return resource


def _make_chain(replication_counter: str) -> list[etree._Element]:
    # A -> B -> C -> D -> E (resptr-prop)

    resources = _make_list_of_resources(5, replication_counter)
    resptr = [_make_resptr_prop(x.attrib["id"]) for x in resources[1:]]
    for i, link in enumerate(resptr):
        resources[i].append(link)
    return resources


def _make_one_circle_with_three_resources(replication_counter: str) -> list[etree._Element]:
    # A -> B (xml-text)
    # A -> B (resptr-prop)
    # B -> C (xml-text)
    # B -> C (resptr-prop)
    # C -> A (xml-text)
    # C -> A (resptr-prop)

    res_li = _make_list_of_resources(number_of_resources=3, replication_counter=replication_counter)
    salsah_list = [_make_xml_text_prop(x) for x in res_li]
    resptr_list = [_make_resptr_prop(x.attrib["id"]) for x in res_li]
    res_li[0].append(salsah_list[1])
    res_li[0].append(resptr_list[1])
    res_li[1].append(salsah_list[2])
    res_li[1].append(resptr_list[2])
    res_li[2].append(salsah_list[0])
    res_li[2].append(resptr_list[0])
    return res_li


def _make_two_references(replication_counter: str) -> list[etree._Element]:
    # A -> B (resptr-prop)
    # B -> A (xml-text)

    res_li = _make_list_of_resources(2, replication_counter)
    res_li[0].append(_make_resptr_prop(res_li[1].attrib["id"]))
    res_li[1].append(_make_xml_text_prop(res_li[0]))
    return res_li


def _make_complex_dependencies_single_link(replication_counter: str) -> list[etree._Element]:
    # A -> D (xml-text)
    # B -> D (xml-text)
    # C -> D (xml-text)
    # D -> E (resptr-prop)
    # E -> A (resptr-prop)
    # E -> B (resptr-prop)
    # E -> C (resptr-prop)

    all_resources = _make_list_of_resources(number_of_resources=5, replication_counter=replication_counter)
    all_resources = _make_complex_dependencies_single_link_resource_ABC(all_resources)
    all_resources = _make_complex_dependencies_single_link_resource_D(all_resources)
    return _make_complex_dependencies_single_link_resource_E(all_resources)


def _make_complex_dependencies_single_link_resource_ABC(resource_list: list[etree._Element]) -> list[etree._Element]:
    link_l = [_make_xml_text_prop(target_res=resource_list[3]) for _ in range(3)]
    for i in range(3):
        resource_list[i].append(link_l[i])
    return resource_list


def _make_complex_dependencies_single_link_resource_D(resource_list: list[etree._Element]) -> list[etree._Element]:
    resource_list[3].append(_make_resptr_prop(resource_list[4].attrib["id"]))
    return resource_list


def _make_complex_dependencies_single_link_resource_E(resource_list: list[etree._Element]) -> list[etree._Element]:
    resource_list[4].append(_make_resptr_prop(target_id=[x.attrib["id"] for x in resource_list[:3]]))
    return resource_list


def _make_complex_dependencies_add_on(replication_counter: str) -> list[etree._Element]:
    # Same as _make_complex_dependencies
    # plus
    # D -> F (xml-text)
    # F -> E (resptr-prop)

    complex_dep_li = _make_complex_dependencies_single_link(replication_counter)
    f_id = f"res_F_{replication_counter}"
    f_res: etree._Element = excel2xml.make_resource(restype=":TestThing", label=f_id, id=f_id)
    s_link = _make_xml_text_prop(target_res=f_res)
    f_res.append(_make_resptr_prop(complex_dep_li[-1].attrib["id"]))
    complex_dep_li[3].append(s_link)
    complex_dep_li.append(f_res)
    return complex_dep_li


def _make_complex_dependencies_with_simpletext(replication_counter: str) -> list[etree._Element]:
    # Same as _make_complex_dependencies
    # plus each value has a simple text property

    complex_dep_li = _make_complex_dependencies_single_link(replication_counter)
    return [_make_simple_text(x) for x in complex_dep_li]


def _make_inverted_complex_dependencies(replication_counter: str) -> list[etree._Element]:
    # A -> D (resptr-prop)
    # B -> D (resptr-prop)
    # C -> D (resptr-prop)
    # D -> E (xml-text)
    # E -> A & B & C (xml-text) single <text> ele

    all_resources = _make_list_of_resources(number_of_resources=5, replication_counter=replication_counter)
    all_resources = _make_single_link_inverted_complex_dependencies_resource_ABC(all_resources)
    all_resources = _make_inverted_complex_dependencies_resource_D(all_resources)
    return _make_inverted_complex_dependencies_resource_E(all_resources)


def _make_single_link_inverted_complex_dependencies_resource_ABC(
    resource_list: list[etree._Element],
) -> list[etree._Element]:
    link_l = [_make_resptr_prop(target_id=resource_list[3].attrib["id"]) for _ in range(3)]
    for i in range(3):
        resource_list[i].append(link_l[i])
    return resource_list


def _make_inverted_complex_dependencies_resource_D(resource_list: list[etree._Element]) -> list[etree._Element]:
    resource_list[3].append(_make_xml_text_prop(resource_list[4]))
    return resource_list


def _make_inverted_complex_dependencies_resource_E(resource_list: list[etree._Element]) -> list[etree._Element]:
    resource_list[4].append(_make_xml_text_prop(target_res=resource_list[:3]))
    return resource_list


def _make_complex_circular_dependencies_multi_link(replication_counter: str) -> list[etree._Element]:
    # Same as _make_complex_dependencies_single_link
    # D -> E (resptr-prop) * 5
    all_resources = _make_list_of_resources(5, replication_counter)
    all_resources = _make_complex_dependencies_single_link_resource_ABC(all_resources)
    all_resources = _make_multi_link_complex_dependencies_resource_D(all_resources)
    return _make_complex_dependencies_single_link_resource_E(all_resources)


def _make_multi_link_complex_dependencies_resource_D(resource_list: list[etree._Element]) -> list[etree._Element]:
    target_resource = resource_list[4]
    links = [_make_resptr_prop(target_resource.attrib["id"], property_name=f":hasResource{n}") for n in range(1, 6)]
    resource_list[3].extend(links)
    return resource_list


def _make_multi_link_two_resource_circle(replication_counter: str) -> list[etree._Element]:
    # A -> B (resptr-prop) * 2
    # B -> A (xml-text)

    res_li = _make_list_of_resources(2, replication_counter)
    links = [_make_resptr_prop(res_li[1].attrib["id"], property_name=f":hasResource{n}") for n in range(1, 3)]
    res_li[0].extend(links)
    res_li[1].append(_make_xml_text_prop(res_li[0]))
    return res_li


def _make_two_resource_circle_plus_non_circle_link(replication_counter: str) -> list[etree._Element]:
    # A -> B (resptr-prop)
    # A -> C (resptr-prop)
    # B -> A (xml-text)

    all_resources = _make_list_of_resources(3, replication_counter)
    all_resources[0].append(_make_resptr_prop([x.attrib["id"] for x in all_resources[1:]]))
    all_resources[1].append(_make_xml_text_prop(all_resources[0]))
    return all_resources


def _make_single_text_ele_for_text_prop(target_resource_id: list[str]) -> etree._Element:
    salsa_links = [_make_salsah_link(x) for x in target_resource_id]
    text_values = [f"Start text{link}end text." for link in salsa_links]
    xml_props = [excel2xml.PropertyElement(text_value, encoding="xml") for text_value in text_values]
    return excel2xml.make_text_prop(name=":hasRichtext", value=xml_props)


def _make_three_resource_circle_with_multiple_text_prop(replication_counter: str) -> list[etree._Element]:
    # A -> B (resptr-prop)
    # A -> C (resptr-prop)
    # C -> A (xml-prop)
    # C -> B (xml-prop)

    all_resources = _make_list_of_resources(3, replication_counter)
    xml_links = _make_single_text_ele_for_text_prop([x.attrib["id"] for x in all_resources[:2]])
    all_resources[2].append(xml_links)
    resptr_links = _make_resptr_prop([x.attrib["id"] for x in all_resources[1:]])
    all_resources[0].append(resptr_links)
    return all_resources


def _make_three_resource_circle_multiple_diverse_links(replication_counter: str) -> list[etree._Element]:
    # A -> B (resptr-prop)
    # A -> B & C (xml-text) single <text> ele
    # A -> C (resptr-prop)
    # C -> A (xml-prop)
    # C -> B (xml-prop)

    resources = _make_three_resource_circle_with_multiple_text_prop(replication_counter)
    resources[0].append(_make_xml_text_prop(resources[1:]))
    return resources


def _make_complex_circle_with_leaf_nodes(replication_counter: str) -> list[etree._Element]:
    # Same as _make_complex_dependencies_add_on
    # plus
    # F ->
    resource_list = _make_complex_dependencies_add_on(replication_counter)
    leaf_resources = _make_list_of_resources(15, replication_counter, 75)
    xml_prop = _make_single_text_ele_for_text_prop([x.attrib["id"] for x in leaf_resources])
    resource_list[-1].append(xml_prop)
    resource_list.extend(leaf_resources)
    return resource_list


if __name__ == "__main__":
    create_and_save_circular_references_test_graph()



from typing import Any

import pytest
import rustworkx as rx
from lxml import etree
from pytest_unordered import unordered

from dsp_tools.commands.xmlupload.stash.construct_and_analyze_graph import _add_stash_to_lookup_dict
from dsp_tools.commands.xmlupload.stash.construct_and_analyze_graph import (
    _create_info_from_xml_for_graph_from_one_resource,
)
from dsp_tools.commands.xmlupload.stash.construct_and_analyze_graph import _create_resptr_link_objects
from dsp_tools.commands.xmlupload.stash.construct_and_analyze_graph import _create_text_link_objects
from dsp_tools.commands.xmlupload.stash.construct_and_analyze_graph import _extract_ids_from_one_text_value
from dsp_tools.commands.xmlupload.stash.construct_and_analyze_graph import _find_cheapest_outgoing_links
from dsp_tools.commands.xmlupload.stash.construct_and_analyze_graph import _find_phantom_xml_edges
from dsp_tools.commands.xmlupload.stash.construct_and_analyze_graph import _remove_edges_to_stash
from dsp_tools.commands.xmlupload.stash.construct_and_analyze_graph import _remove_leaf_nodes
from dsp_tools.commands.xmlupload.stash.construct_and_analyze_graph import create_info_from_xml_for_graph
from dsp_tools.commands.xmlupload.stash.construct_and_analyze_graph import generate_upload_order
from dsp_tools.commands.xmlupload.stash.construct_and_analyze_graph import make_graph
from dsp_tools.commands.xmlupload.stash.graph_models import Edge
from dsp_tools.commands.xmlupload.stash.graph_models import ResptrLink
from dsp_tools.commands.xmlupload.stash.graph_models import XMLLink


def test_create_info_from_xml_for_graph_from_one_resource() -> None:
    test_ele = etree.fromstring(
        """<resource label="res_A_19" restype=":TestThing" id="res_A_19" permissions="open">
            <resptr-prop name=":hasResource1">
                <resptr permissions="open">res_B_19</resptr>
                <resptr permissions="open">res_C_19</resptr>
            </resptr-prop>
            <text-prop name=":hasRichtext">
                <text permissions="open" encoding="xml">
                    <a class="salsah-link" href="IRI:res_B_19:IRI">res_B_19</a>
                    <a class="salsah-link" href="IRI:res_C_19:IRI">res_C_19</a>
                </text>
            </text-prop>
        </resource>"""
    )
    res_resptr_links, res_xml_links = _create_info_from_xml_for_graph_from_one_resource(test_ele)
    res_B_19 = [obj.target_id for obj in res_resptr_links]
    assert "res_B_19" in res_B_19
    assert "res_C_19" in res_B_19
    assert res_xml_links[0].source_id == "res_A_19"
    assert res_xml_links[0].target_ids == {"res_B_19", "res_C_19"}


def test_create_info_from_xml_for_graph_from_one_resource_one() -> None:
    test_ele = etree.fromstring(
        """<resource label="res_A_11" restype=":TestThing" id="res_A_11" permissions="open">
            <text-prop name=":hasRichtext">
                <text permissions="open" encoding="xml">
                    <a class="salsah-link" href="IRI:res_B_11:IRI">res_B_11</a>
                </text>
            </text-prop>
            <resptr-prop name=":hasResource1">
                <resptr permissions="open">res_B_11</resptr>
            </resptr-prop>
        </resource>"""
    )
    res_resptr, res_xml = _create_info_from_xml_for_graph_from_one_resource(test_ele)
    assert res_resptr[0].target_id == "res_B_11"
    assert isinstance(res_resptr[0], ResptrLink)
    assert res_xml[0].target_ids == {"res_B_11"}
    assert isinstance(res_xml[0], XMLLink)


def test_create_info_from_xml_for_graph_from_one_resource_no_links() -> None:
    test_ele = etree.fromstring('<resource label="res_B_18" restype=":TestThing" id="res_B_18" permissions="open"/>')
    res_resptr, res_xml = _create_info_from_xml_for_graph_from_one_resource(test_ele)
    assert (res_resptr, res_xml) == ([], [])


def test_text_only_create_info_from_xml_for_graph_from_one_resource() -> None:
    test_ele = etree.fromstring(
        """<resource label="res_C_18" restype=":TestThing" id="res_C_18" permissions="open">
            <text-prop name=":hasRichtext">
                <text permissions="open" encoding="xml">
                    <a class="salsah-link" href="IRI:res_A_18:IRI">res_A_18</a>
                </text>
                <text permissions="open" encoding="xml">
                    <a class="salsah-link" href="IRI:res_B_18:IRI">res_B_18</a>
                </text>
            </text-prop>
        </resource>"""
    )
    res_resptr, res_xml = _create_info_from_xml_for_graph_from_one_resource(test_ele)
    assert not res_resptr
    res_xml_ids = [x.target_ids for x in res_xml]
    assert unordered(res_xml_ids) == [{"res_A_18"}, {"res_B_18"}]


def test_extract_id_one_text_with_one_id() -> None:
    test_ele = etree.fromstring(
        """<text permissions="open" encoding="xml">
            <a class="salsah-link" href="IRI:res_A_11:IRI">res_A_11</a>
        </text>"""
    )
    res = _extract_ids_from_one_text_value(test_ele)
    assert res == {"res_A_11"}


def test_extract_id_one_text_with_iri() -> None:
    test_ele = etree.fromstring(
        '<text permissions="open" encoding="xml">'
        '<a class="salsah-link" href="http://rdfh.ch/0801/RDE7_KU1STuDhHnGr5uu0g">res_A_11</a></text>'
    )
    res = _extract_ids_from_one_text_value(test_ele)
    assert res == set()


def test_extract_id_one_text_with_several_id() -> None:
    test_ele = etree.fromstring(
        """<text permissions="open" encoding="xml">
            <a class="salsah-link" href="IRI:res_A_11:IRI">res_A_11</a>
            <a class="salsah-link" href="IRI:res_B_11:IRI">res_B_11</a>
            <a class="salsah-link" href="IRI:res_B_11:IRI">res_B_11</a>
        </text>"""
    )
    res = _extract_ids_from_one_text_value(test_ele)
    assert res == {"res_A_11", "res_B_11"}


def test_extract_ids_from_text_prop_with_several_text_links() -> None:
    test_ele = etree.fromstring(
        """<text-prop name=":hasRichtext">
            <text permissions="open" encoding="xml">
                <a class="salsah-link" href="IRI:res_A_18:IRI">res_A_18</a>
            </text>
            <text permissions="open" encoding="xml">
                <a class="salsah-link" href="IRI:res_B_18:IRI">res_B_18</a>
            </text>
        </text-prop>"""
    )
    res = _create_text_link_objects("res_C_18", test_ele)
    res_ids = [x.target_ids for x in res]
    assert unordered(res_ids) == [{"res_A_18"}, {"res_B_18"}]


def test_extract_ids_from_text_prop_with_iris_and_ids() -> None:
    test_ele = etree.fromstring(
        """<text-prop name=":hasRichtext">
            <text permissions="open" encoding="xml">
                <a class="salsah-link" href="http://rdfh.ch/4123/vEpjk7zAQBC2j3pvTGSxcw">foo</a>
            </text>
            <text permissions="open" encoding="xml">
                <a class="salsah-link" href="IRI:res_B_18:IRI">res_B_18</a>
            </text>
        </text-prop>"""
    )
    res = _create_text_link_objects("foo", test_ele)
    assert len(res) == 1
    assert res[0].target_ids == {"res_B_18"}
    children = list(test_ele.iterchildren())
    assert not children[0].attrib.get("linkUUID")
    assert children[1].attrib.get("linkUUID")


def test_create_class_instance_resptr_link_one_link() -> None:
    test_ele = etree.fromstring(
        """<resptr-prop xmlns="https://dasch.swiss/schema" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" 
        name=":hasResource1">
            <resptr permissions="open">res_C_15</resptr>
        </resptr-prop>"""
    )
    res = _create_resptr_link_objects("res_A_15", test_ele)
    assert res[0].target_id == "res_C_15"


def test_create_class_instance_resptr_link_several() -> None:
    test_ele = etree.fromstring(
        """<resptr-prop name=":hasResource1">
            <resptr permissions="open">res_A_13</resptr>
            <resptr permissions="open">res_B_13</resptr>
            <resptr permissions="open">res_C_13</resptr>
        </resptr-prop>"""
    )
    res = _create_resptr_link_objects("res_D_13", test_ele)
    assert all(isinstance(x, ResptrLink) for x in res)
    assert res[0].target_id == "res_A_13"
    assert res[1].target_id == "res_B_13"
    assert res[2].target_id == "res_C_13"


def test_create_class_instance_resptr_link_with_iris() -> None:
    test_ele = etree.fromstring(
        """<resptr-prop name=":hasResource1">
            <resptr permissions="open">res_A_13</resptr>
            <resptr permissions="open">res_B_13</resptr>
            <resptr permissions="open">http://rdfh.ch/4123/vEpjk7zAQBC2j3pvTGSxcw</resptr>
        </resptr-prop>"""
    )
    res = _create_resptr_link_objects("foo", test_ele)
    assert len(res) == 2
    assert res[0].target_id == "res_A_13"
    assert res[1].target_id == "res_B_13"
    children = list(test_ele.iterchildren())
    assert children[0].attrib.get("linkUUID")
    assert children[1].attrib.get("linkUUID")
    assert not children[2].attrib.get("linkUUID")


def test_create_info_from_xml_for_graph_check_UUID_in_root() -> None:
    root = etree.fromstring(
        """<knora shortcode="0700" default-ontology="simcir">
            <resource label="res_A_11" restype=":TestThing" id="res_A_11" permissions="open">
                <resptr-prop name=":hasResource1">
                    <resptr permissions="open">res_B_11</resptr>
                </resptr-prop>
            </resource>
            <resource label="res_B_11" restype=":TestThing" id="res_B_11" permissions="open">
                <text-prop name=":hasRichtext">
                    <text permissions="open" encoding="xml">
                        Start text<a class="salsah-link" href="IRI:res_C_11:IRI">res_C_11</a>end text.
                    </text>
                </text-prop>
            </resource>
            <resource label="res_C_11" restype=":TestThing" id="res_C_11" permissions="open"></resource>
        </knora>"""
    )
    res_resptr_li, res_xml_li, res_all_ids = create_info_from_xml_for_graph(root)
    res_resptr = res_resptr_li[0]
    assert isinstance(res_resptr, ResptrLink)
    res_xml = res_xml_li[0]
    assert isinstance(res_xml, XMLLink)
    assert unordered(res_all_ids) == ["res_A_11", "res_B_11", "res_C_11"]
    xml_res_resptr = root.find(".//resptr")
    assert xml_res_resptr.attrib["linkUUID"] == res_resptr.link_uuid  # type: ignore[union-attr]
    xml_res_text = root.find(".//text")
    assert xml_res_text.attrib["linkUUID"] == res_xml.link_uuid  # type: ignore[union-attr]


def test_make_graph() -> None:
    resptr = ResptrLink("a", "b")
    resptr_links = [resptr]
    xml = XMLLink("a", {"b", "c"})
    xml_links = [xml]
    all_ids = ["a", "b", "c"]
    graph, node_to_id, edges = make_graph(resptr_links, xml_links, all_ids)
    assert graph.num_nodes() == 3
    assert graph.num_edges() == 3
    assert node_to_id[0] == "a"
    assert node_to_id[1] == "b"
    assert node_to_id[2] == "c"
    assert unordered(edges) == [Edge(0, 1, resptr), Edge(0, 1, xml), Edge(0, 2, xml)]
    assert set(node_to_id.keys()) == {0, 1, 2}


def test_remove_leaf_nodes() -> None:
    graph: rx.PyDiGraph[Any, Any] = rx.PyDiGraph()
    nodes = ["a", "b", "c", "d", "e", "f"]
    node_idx = set(graph.add_nodes_from(nodes))
    node_idx_lookup = dict(zip(node_idx, nodes))
    graph.add_edges_from(
        [
            (0, 1, "ab"),
            (0, 4, "ae"),
            (0, 2, "ac"),
            (1, 2, "bc"),
            (1, 3, "bd"),
            (2, 4, "ce"),
            (3, 1, "da"),
            (3, 2, "dc"),
            (3, 4, "de"),
        ]
    )
    # c is a second degree leaf
    # e is a leaf
    # f has no edges

    removed_leaf_nodes, remaining_node_indices = _remove_leaf_nodes(graph, node_idx_lookup, node_idx)
    assert unordered(removed_leaf_nodes) == ["c", "e", "f"]
    assert remaining_node_indices == {0, 1, 3}
    assert unordered(graph.nodes()) == ["a", "b", "d"]
    assert unordered(graph.edges()) == ["ab", "bd", "da"]


def test_find_cheapest_outgoing_links_one_resptr_link() -> None:
    nodes = [
        #     out / in
        "0",  # 3 / 3
        "1",  # 2 / 3
        "2",  # 3 / 2
        "3",  # 4 / 2
        "4",
    ]
    graph: rx.PyDiGraph[Any, Any] = rx.PyDiGraph()
    graph.add_nodes_from(nodes)
    circle = [(0, 1), (1, 2), (2, 3), (3, 0)]
    edges = [
        Edge(0, 1, ResptrLink("", "")),
        Edge(0, 4, ResptrLink("", "")),
        Edge(0, 4, ResptrLink("", "")),
        Edge(1, 2, ResptrLink("", "")),
        Edge(1, 3, ResptrLink("", "")),
        Edge(2, 0, ResptrLink("", "")),
        Edge(2, 1, ResptrLink("", "")),
        Edge(2, 3, ResptrLink("", "")),
        Edge(3, 0, ResptrLink("", "")),
        Edge(3, 0, ResptrLink("", "")),
        Edge(3, 1, ResptrLink("", "")),
        Edge(3, 2, ResptrLink("", "")),
    ]
    graph.add_edges_from([e.as_tuple() for e in edges])
    cheapest_links = _find_cheapest_outgoing_links(graph, circle, edges)
    assert cheapest_links == [edges[3]]


def test_find_cheapest_outgoing_links_four_circle() -> None:
    nodes = [
        #     out / in
        "0",  # 1 / 3
        "1",  # 2 / 1
        "2",  # 3 / 6
        "3",  # 6 / 3
        "4",
        "5",
    ]
    graph: rx.PyDiGraph[Any, Any] = rx.PyDiGraph()
    graph.add_nodes_from(nodes)
    edges = [
        Edge(0, 1, ResptrLink("", "")),
        Edge(1, 0, ResptrLink("", "")),
        Edge(1, 2, ResptrLink("", "")),
        Edge(1, 2, ResptrLink("", "")),
        Edge(2, 3, ResptrLink("", "")),
        Edge(2, 3, ResptrLink("", "")),
        Edge(2, 3, ResptrLink("", "")),
        Edge(3, 0, ResptrLink("", "")),
        Edge(3, 0, ResptrLink("", "")),
        Edge(3, 5, ResptrLink("", "")),
        Edge(3, 5, ResptrLink("", "")),
        Edge(3, 5, ResptrLink("", "")),
        Edge(3, 5, ResptrLink("", "")),
        Edge(4, 2, ResptrLink("", "")),
        Edge(4, 2, ResptrLink("", "")),
        Edge(4, 2, ResptrLink("", "")),
        Edge(4, 2, ResptrLink("", "")),
    ]
    graph.add_edges_from([e.as_tuple() for e in edges])
    circle = [(0, 1), (1, 2), (2, 3), (3, 0)]
    cheapest_links = _find_cheapest_outgoing_links(graph, circle, edges)
    assert cheapest_links == [edges[0]]


def test_find_cheapest_outgoing_links_xml() -> None:
    nodes = [
        #      out / in
        "0",  # 4 (2 XML) / 3
        "1",  # 3 / 3
        "2",  # 1 (3 XML) / 3
        "3",  # 3 / 3
        "4",
        "5",
    ]
    graph: rx.PyDiGraph[Any, Any] = rx.PyDiGraph()
    graph.add_nodes_from(nodes)
    a_de_xml = XMLLink("0", {"3", "4"})
    b_d_xml = XMLLink("1", {"3"})
    c_bdf_xml = XMLLink("2", {"1", "3", "5"})
    circle = [(0, 1), (1, 2), (2, 3), (3, 0)]
    edges = [
        Edge(0, 1, ResptrLink("", "")),
        Edge(0, 1, ResptrLink("", "")),
        Edge(0, 2, ResptrLink("", "")),
        Edge(1, 2, ResptrLink("", "")),
        Edge(1, 2, ResptrLink("", "")),
        Edge(3, 0, ResptrLink("", "")),
        Edge(3, 0, ResptrLink("", "")),
        Edge(3, 0, ResptrLink("", "")),
        Edge(1, 3, b_d_xml),
        Edge(2, 1, c_bdf_xml),
        Edge(2, 3, c_bdf_xml),
        Edge(2, 5, c_bdf_xml),
        Edge(0, 3, a_de_xml),
        Edge(0, 4, a_de_xml),
    ]
    graph.add_edges_from([e.as_tuple() for e in edges])
    cheapest_links = _find_cheapest_outgoing_links(graph, circle, edges)
    assert cheapest_links == [edges[10]]


def test_remove_edges_to_stash_phantom_xml() -> None:
    nodes = ["0", "1", "2", "3", "4", "5"]
    graph: rx.PyDiGraph[Any, Any] = rx.PyDiGraph()
    graph.add_nodes_from(nodes)
    a_de_xml = XMLLink("0", {"3", "4"})
    b_d_xml = XMLLink("1", {"3"})
    c_bdf_xml = XMLLink("2", {"1", "3", "5"})
    edges_to_remove = [Edge(2, 3, c_bdf_xml)]
    remaining_nodes = set(range(10))
    edges = [
        Edge(0, 1, ResptrLink("", "")),
        Edge(0, 1, ResptrLink("", "")),
        Edge(0, 2, ResptrLink("", "")),
        Edge(1, 2, ResptrLink("", "")),
        Edge(1, 2, ResptrLink("", "")),
        Edge(3, 0, ResptrLink("", "")),
        Edge(3, 0, ResptrLink("", "")),
        Edge(3, 0, ResptrLink("", "")),
        Edge(0, 3, a_de_xml),
        Edge(0, 4, a_de_xml),
        Edge(1, 3, b_d_xml),
        Edge(2, 1, c_bdf_xml),
        Edge(2, 3, c_bdf_xml),
        Edge(2, 5, c_bdf_xml),
    ]
    graph.add_edges_from([e.as_tuple() for e in edges])
    _remove_edges_to_stash(graph, edges_to_remove, edges, remaining_nodes)
    remaining_edges = list(graph.edge_list())
    expected_edges = [(0, 1), (0, 1), (0, 2), (0, 3), (0, 4), (1, 2), (1, 2), (1, 3), (3, 0), (3, 0), (3, 0)]
    assert unordered(remaining_edges) == expected_edges


def test_remove_edges_to_stash_several_resptr() -> None:
    nodes = ["0", "1", "2"]
    graph: rx.PyDiGraph[Any, Any] = rx.PyDiGraph()
    graph.add_nodes_from(nodes)
    edges = [
        Edge(0, 1, ResptrLink("", "")),
        Edge(0, 1, ResptrLink("", "")),
        Edge(1, 2, ResptrLink("", "")),
        Edge(1, 2, ResptrLink("", "")),
        Edge(1, 2, ResptrLink("", "")),
        Edge(1, 2, ResptrLink("", "")),
        Edge(1, 2, ResptrLink("", "")),
        Edge(2, 0, ResptrLink("", "")),
        Edge(2, 0, ResptrLink("", "")),
        Edge(2, 0, ResptrLink("", "")),
        Edge(2, 0, ResptrLink("", "")),
    ]
    graph.add_edges_from([e.as_tuple() for e in edges])
    edges_to_remove = edges[:2]
    remaining_nodes = set(range(10))
    _remove_edges_to_stash(graph, edges_to_remove, edges, remaining_nodes)
    remaining_edges = list(graph.edge_list())
    assert unordered(remaining_edges) == [(1, 2), (1, 2), (1, 2), (1, 2), (1, 2), (2, 0), (2, 0), (2, 0), (2, 0)]


def test_remove_edges_to_stash_missing_nodes() -> None:
    nodes = ["a", "b", "c", "d"]
    graph: rx.PyDiGraph[Any, Any] = rx.PyDiGraph()
    graph.add_nodes_from(nodes)
    xml_link = XMLLink("a", {"b", "d"})
    edges = [
        Edge(0, 1, xml_link),
        Edge(1, 2, ResptrLink("", "")),
        Edge(2, 0, ResptrLink("", "")),
    ]
    graph.add_edges_from([e.as_tuple() for e in edges])
    remaining_nodes = {0, 1, 2}
    edges_to_remove = [Edge(0, 1, xml_link)]
    _remove_edges_to_stash(graph, edges_to_remove, edges, remaining_nodes)
    remaining_edges = list(graph.edge_list())
    assert unordered(remaining_edges) == [(1, 2), (2, 0)]


def test_find_phantom_xml_edges_no_remaining() -> None:
    xml_link = XMLLink("0", {"2", "3"})
    edges = [
        Edge(0, 1, xml_link),
        Edge(1, 2, ResptrLink("", "")),
        Edge(2, 0, ResptrLink("", "")),
    ]
    remaining_nodes = {0, 1, 2}
    phantoms = _find_phantom_xml_edges(0, 1, edges, xml_link, remaining_nodes)
    assert phantoms == []


def test_find_phantom_xml_edges_one_link() -> None:
    xml_link = XMLLink("0", {"1", "3"})
    edges = [
        Edge(0, 1, xml_link),
        Edge(0, 3, xml_link),
        Edge(1, 2, ResptrLink("", "")),
        Edge(2, 0, ResptrLink("", "")),
    ]
    remaining_nodes = {0, 1, 2, 3}
    phantoms = _find_phantom_xml_edges(0, 1, edges, xml_link, remaining_nodes)
    assert phantoms == [(0, 3)]


def test_add_stash_to_lookup_dict_none_existing() -> None:
    resptr_a1 = ResptrLink("0", "1")
    resptr_a2 = ResptrLink("0", "1")
    xml_a = XMLLink("0", {"1", "2"})
    to_stash: list[XMLLink | ResptrLink] = [resptr_a1, resptr_a2, xml_a]
    expected = {"0": [resptr_a1.link_uuid, resptr_a2.link_uuid, xml_a.link_uuid]}
    result = _add_stash_to_lookup_dict({}, to_stash)
    assert result.keys() == expected.keys()
    assert unordered(result["0"]) == expected["0"]


def test_add_stash_to_lookup_dict() -> None:
    resptr_a1 = ResptrLink("0", "1")
    stash_dict = {"0": ["existingUUID1", "existingUUID2"], "1": ["existingUUID1"]}
    expected = {"0": ["existingUUID1", "existingUUID2", resptr_a1.link_uuid], "1": ["existingUUID1"]}
    result = _add_stash_to_lookup_dict(stash_dict, [resptr_a1])
    assert result.keys() == expected.keys()
    assert unordered(result["0"]) == expected["0"]


def test_generate_upload_order_with_stash() -> None:
    graph: rx.PyDiGraph[Any, Any] = rx.PyDiGraph()
    nodes = ["0", "1", "2", "3", "4", "5", "6"]
    node_idx = set(graph.add_nodes_from(nodes))
    node_idx_lookup = dict(zip(node_idx, nodes))
    abf_xml = XMLLink("0", {"1", "5"})
    edges = [
        Edge(1, 2, ResptrLink("", "")),
        Edge(2, 3, ResptrLink("", "")),
        Edge(3, 0, ResptrLink("", "")),
        Edge(3, 0, ResptrLink("", "")),
        Edge(3, 0, ResptrLink("", "")),
        Edge(3, 4, ResptrLink("", "")),
        Edge(5, 6, ResptrLink("", "")),
        Edge(0, 1, abf_xml),
        Edge(0, 5, abf_xml),
    ]
    graph.add_edges_from([e.as_tuple() for e in edges])
    stash_lookup, upload_order, stash_counter = generate_upload_order(graph, node_idx_lookup, edges)
    expected_stash_lookup = {"0": [abf_xml.link_uuid]}
    assert stash_counter == 1
    assert unordered(upload_order[:2]) == ["4", "6"]
    assert upload_order[2:] == ["5", "0", "3", "2", "1"]
    assert stash_lookup.keys() == expected_stash_lookup.keys()
    assert stash_lookup["0"] == expected_stash_lookup["0"]
    assert not list(graph.edges())
    assert not list(graph.nodes())


def test_generate_upload_order_no_stash() -> None:
    graph: rx.PyDiGraph[Any, Any] = rx.PyDiGraph()
    nodes = ["0", "1", "2", "3"]
    node_idx = set(graph.add_nodes_from(nodes))
    node_idx_lookup = dict(zip(node_idx, nodes))
    edges = [
        Edge(0, 1, ResptrLink("", "")),
        Edge(1, 2, ResptrLink("", "")),
        Edge(2, 3, ResptrLink("", "")),
    ]
    graph.add_edges_from([e.as_tuple() for e in edges])
    stash_lookup, upload_order, stash_counter = generate_upload_order(graph, node_idx_lookup, edges)
    assert not stash_lookup
    assert stash_counter == 0
    assert upload_order == ["3", "2", "1", "0"]
    assert not list(graph.edges())
    assert not list(graph.nodes())


def test_generate_upload_order_two_circles() -> None:
    graph: rx.PyDiGraph[Any, Any] = rx.PyDiGraph()
    nodes = ["0", "1", "2", "3", "4", "5", "6"]
    node_idx = set(graph.add_nodes_from(nodes))
    node_idx_lookup = dict(zip(node_idx, nodes))
    edges = [
        Edge(0, 1, ResptrLink("0", "1")),
        Edge(0, 5, ResptrLink("0", "5")),
        Edge(1, 2, ResptrLink("1", "2")),
        Edge(2, 3, ResptrLink("2", "3")),
        Edge(2, 0, ResptrLink("2", "0")),
        Edge(3, 0, ResptrLink("3", "0")),
        Edge(3, 0, ResptrLink("3", "0")),
        Edge(3, 0, ResptrLink("3", "0")),
        Edge(3, 4, ResptrLink("3", "4")),
        Edge(5, 6, ResptrLink("5", "6")),
        Edge(5, 6, ResptrLink("5", "6")),
        Edge(6, 5, ResptrLink("6", "5")),
        Edge(6, 5, ResptrLink("6", "5")),
        Edge(6, 5, ResptrLink("6", "5")),
    ]
    graph.add_edges_from([e.as_tuple() for e in edges])
    stash_lookup, upload_order, stash_counter = generate_upload_order(graph, node_idx_lookup, edges)
    circles = ["0", "1", "2", "3", "5", "6"]
    expected_stash = {"0": [edges[0].link_object.link_uuid], "5": [x.link_object.link_uuid for x in edges[9:11]]}
    assert upload_order[0] == "4"
    assert unordered(upload_order[1:]) == circles
    assert stash_counter == 3
    assert stash_lookup.keys() == expected_stash.keys()
    assert stash_lookup["0"] == expected_stash["0"]
    assert unordered(stash_lookup["5"]) == expected_stash["5"]
    assert not list(graph.edges())
    assert not list(graph.nodes())


if __name__ == "__main__":
    pytest.main([__file__])



from copy import deepcopy
from dataclasses import dataclass
from typing import Any
from typing import cast
from unittest.mock import Mock
from unittest.mock import patch

import pytest
from lxml import etree
from requests import Response

from dsp_tools.commands.xmlupload.iri_resolver import IriResolver
from dsp_tools.commands.xmlupload.models.deserialise.xmlresource import XMLResource
from dsp_tools.commands.xmlupload.models.ingest import AssetClient
from dsp_tools.commands.xmlupload.models.ingest import DspIngestClientLive
from dsp_tools.commands.xmlupload.models.upload_clients import UploadClients
from dsp_tools.commands.xmlupload.models.upload_state import UploadState
from dsp_tools.commands.xmlupload.project_client import ProjectInfo
from dsp_tools.commands.xmlupload.stash.stash_models import LinkValueStash
from dsp_tools.commands.xmlupload.stash.stash_models import LinkValueStashItem
from dsp_tools.commands.xmlupload.stash.stash_models import Stash
from dsp_tools.commands.xmlupload.upload_config import UploadConfig
from dsp_tools.commands.xmlupload.xmlupload import _upload_resources
from dsp_tools.models.custom_warnings import DspToolsUserWarning
from dsp_tools.models.exceptions import PermanentTimeOutError
from dsp_tools.models.exceptions import XmlUploadInterruptedError
from dsp_tools.utils.connection import Connection
from dsp_tools.utils.connection_live import ConnectionLive
from test.integration.commands.xmlupload.authentication_client_mock import AuthenticationClientMockBase


@pytest.fixture
def ingest_client_mock():  # type: ignore[no-untyped-def]
    return Mock(spec_set=AssetClient)


class ListClientMock:
    def get_list_node_id_to_iri_lookup(self) -> dict[str, str]:
        return {}


@dataclass
class ProjectClientStub:
    """Stub class for ProjectClient."""

    con: Connection
    shortcode: str
    project_info: ProjectInfo | None

    def get_project_iri(self) -> str:
        return "https://admin.test.dasch.swiss/project/MsOaiQkcQ7-QPxsYBKckfQ"

    def get_ontology_iris(self) -> list[str]:
        raise NotImplementedError("get_project_iri not implemented")

    def get_ontology_name_dict(self) -> dict[str, str]:
        return {"my_onto": "http://0.0.0.0:3333/ontology/9999/onto/v2"}

    def get_ontology_iri_dict(self) -> dict[str, str]:
        raise NotImplementedError("get_project_iri not implemented")


def test_one_resource_without_links(ingest_client_mock: AssetClient) -> None:
    xml_strings = [
        """
        <resource label="foo_1_label" restype=":foo_1_type" id="foo_1_id">
            <text-prop name=":hasSimpleText"><text encoding="utf8">foo_1 text</text></text-prop>
        </resource>
        """,
    ]
    xml_resources = [XMLResource.from_node(etree.fromstring(xml_str), "my_onto") for xml_str in xml_strings]
    upload_state = UploadState(xml_resources, None, UploadConfig(), {})
    con = Mock(spec_set=ConnectionLive)
    post_responses = [{"@id": "foo_1_iri", "rdfs:label": "foo_1_label"}]
    con.post = Mock(side_effect=post_responses)
    project_client = ProjectClientStub(con, "1234", None)

    clients = UploadClients(ingest_client_mock, project_client, ListClientMock())
    _upload_resources(clients, upload_state)

    assert len(con.post.call_args_list) == len(post_responses)
    post_call_args = cast(dict[str, Any], con.post.call_args_list[0].kwargs)
    assert len(post_call_args) == 3
    prop_name = "http://0.0.0.0:3333/ontology/9999/onto/v2#hasSimpleText"
    expected = {
        "route": "/v2/resources",
        "data": {
            "@type": "http://0.0.0.0:3333/ontology/9999/onto/v2#foo_1_type",
            "http://www.w3.org/2000/01/rdf-schema#label": "foo_1_label",
            "http://api.knora.org/ontology/knora-api/v2#attachedToProject": {
                "@id": "https://admin.test.dasch.swiss/project/MsOaiQkcQ7-QPxsYBKckfQ"
            },
            "@context": {},
            prop_name: {
                "@type": "http://api.knora.org/ontology/knora-api/v2#TextValue",
                "http://api.knora.org/ontology/knora-api/v2#valueAsString": "foo_1 text",
            },
        },
    }
    assert post_call_args["route"] == expected["route"]
    assert not post_call_args["headers"]
    assert post_call_args["data"][prop_name] == expected["data"][prop_name]  # type: ignore[index]
    expected_project = expected["data"]["http://api.knora.org/ontology/knora-api/v2#attachedToProject"]  # type: ignore[index]
    assert post_call_args["data"]["http://api.knora.org/ontology/knora-api/v2#attachedToProject"] == expected_project
    expected_label = expected["data"]["http://www.w3.org/2000/01/rdf-schema#label"]  # type: ignore[index]
    assert post_call_args["data"]["http://www.w3.org/2000/01/rdf-schema#label"] == expected_label
    assert post_call_args["data"]["@type"] == expected["data"]["@type"]  # type: ignore[index]
    assert not upload_state.pending_resources
    assert not upload_state.failed_uploads
    assert upload_state.iri_resolver.lookup == {"foo_1_id": "foo_1_iri"}
    assert not upload_state.pending_stash


def test_one_resource_with_link_to_existing_resource(ingest_client_mock: AssetClient) -> None:
    xml_strings = [
        """
        <resource label="foo_1_label" restype=":foo_1_type" id="foo_1_id">
            <resptr-prop name=":hasCustomLink"><resptr>foo_2_id</resptr></resptr-prop>
        </resource>
        """,
    ]
    xml_resources = [XMLResource.from_node(etree.fromstring(xml_str), "my_onto") for xml_str in xml_strings]
    upload_state = UploadState(xml_resources, None, UploadConfig(), {}, [], IriResolver({"foo_2_id": "foo_2_iri"}))
    con = Mock(spec_set=ConnectionLive)
    post_responses = [{"@id": "foo_1_iri", "rdfs:label": "foo_1_label"}]
    con.post = Mock(side_effect=post_responses)
    project_client = ProjectClientStub(con, "1234", None)
    clients = UploadClients(ingest_client_mock, project_client, ListClientMock())
    _upload_resources(clients, upload_state)

    assert len(con.post.call_args_list) == len(post_responses)
    post_call_args = cast(dict[str, Any], con.post.call_args_list[0].kwargs)
    assert len(post_call_args) == 3
    prop_name = "http://0.0.0.0:3333/ontology/9999/onto/v2#hasCustomLinkValue"
    expected = {
        "route": "/v2/resources",
        "data": {
            "@type": "http://0.0.0.0:3333/ontology/9999/onto/v2#foo_1_type",
            "http://www.w3.org/2000/01/rdf-schema#label": "foo_1_label",
            "http://api.knora.org/ontology/knora-api/v2#attachedToProject": {
                "@id": "https://admin.test.dasch.swiss/project/MsOaiQkcQ7-QPxsYBKckfQ"
            },
            "@context": {},
            prop_name: {
                "@type": "http://api.knora.org/ontology/knora-api/v2#LinkValue",
                "http://api.knora.org/ontology/knora-api/v2#linkValueHasTargetIri": {"@id": "foo_2_iri"},
            },
        },
    }
    assert post_call_args["route"] == expected["route"]
    assert not post_call_args["headers"]
    assert post_call_args["data"][prop_name] == expected["data"][prop_name]  # type: ignore[index]
    expected_project = expected["data"]["http://api.knora.org/ontology/knora-api/v2#attachedToProject"]  # type: ignore[index]
    assert post_call_args["data"]["http://api.knora.org/ontology/knora-api/v2#attachedToProject"] == expected_project
    expected_label = expected["data"]["http://www.w3.org/2000/01/rdf-schema#label"]  # type: ignore[index]
    assert post_call_args["data"]["http://www.w3.org/2000/01/rdf-schema#label"] == expected_label
    assert post_call_args["data"]["@type"] == expected["data"]["@type"]  # type: ignore[index]
    assert not upload_state.pending_resources
    assert not upload_state.failed_uploads
    assert upload_state.iri_resolver.lookup == {"foo_1_id": "foo_1_iri", "foo_2_id": "foo_2_iri"}
    assert not upload_state.pending_stash


def test_2_resources_with_stash_interrupted_by_timeout(ingest_client_mock: AssetClient) -> None:
    _2_resources_with_stash_interrupted_by_error(PermanentTimeOutError(""), "PermanentTimeOutError", ingest_client_mock)


def test_2_resources_with_stash_interrupted_by_keyboard(ingest_client_mock: AssetClient) -> None:
    _2_resources_with_stash_interrupted_by_error(KeyboardInterrupt(), "KeyboardInterrupt", ingest_client_mock)


def _2_resources_with_stash_interrupted_by_error(
    err_to_interrupt_with: BaseException, err_as_str: str, ingest_client_mock: AssetClient
) -> None:
    xml_strings = [
        '<resource label="foo_1_label" restype=":foo_1_type" id="foo_1_id"></resource>',
        '<resource label="foo_2_label" restype=":foo_2_type" id="foo_2_id"></resource>',
    ]
    xml_resources = [XMLResource.from_node(etree.fromstring(xml_str), "my_onto") for xml_str in xml_strings]
    link_val_stash_dict = {
        "foo_1_id": [LinkValueStashItem("foo_1_id", "my_onto:foo_1_type", "my_onto:hasCustomLink", "foo_2_id")],
        "foo_2_id": [LinkValueStashItem("foo_2_id", "my_onto:foo_2_type", "my_onto:hasCustomLink", "foo_1_id")],
    }
    stash = Stash(link_value_stash=LinkValueStash(link_val_stash_dict), standoff_stash=None)
    upload_state = UploadState(xml_resources.copy(), deepcopy(stash), UploadConfig(), {})
    con = Mock(spec_set=ConnectionLive)
    post_responses = [
        {"@id": "foo_1_iri", "rdfs:label": "foo_1_label"},
        err_to_interrupt_with,
    ]
    con.post = Mock(side_effect=post_responses)
    project_client = ProjectClientStub(con, "1234", None)
    clients = UploadClients(ingest_client_mock, project_client, ListClientMock())

    with patch("dsp_tools.commands.xmlupload.xmlupload._handle_upload_error") as _handle_upload_error:
        with pytest.warns(DspToolsUserWarning):
            _upload_resources(clients, upload_state)

        assert len(con.post.call_args_list) == len(post_responses)
        err_msg = (
            f"There was a {err_as_str} while trying to create resource 'foo_2_id'.\n"
            "It is unclear if the resource 'foo_2_id' was created successfully or not.\n"
            "Please check manually in the DSP-APP or DB.\n"
            "In case of successful creation, call 'resume-xmlupload' with the flag "
            "'--skip-first-resource' to prevent duplication.\n"
            "If not, a normal 'resume-xmlupload' can be started."
        )
        upload_state_expected = UploadState(
            xml_resources[1:], stash, UploadConfig(), {}, [], IriResolver({"foo_1_id": "foo_1_iri"})
        )
        _handle_upload_error.assert_called_once_with(XmlUploadInterruptedError(err_msg), upload_state_expected)


def test_2_resources_with_stash(ingest_client_mock: AssetClient) -> None:
    xml_strings = [
        '<resource label="foo_1_label" restype=":foo_1_type" id="foo_1_id"></resource>',
        '<resource label="foo_2_label" restype=":foo_2_type" id="foo_2_id"></resource>',
    ]
    xml_resources = [XMLResource.from_node(etree.fromstring(xml_str), "my_onto") for xml_str in xml_strings]
    link_val_stash_dict = {
        "foo_1_id": [LinkValueStashItem("foo_1_id", "my_onto:foo_1_type", "my_onto:hasCustomLink", "foo_2_id")],
        "foo_2_id": [LinkValueStashItem("foo_2_id", "my_onto:foo_2_type", "my_onto:hasCustomLink", "foo_1_id")],
    }
    stash = Stash(link_value_stash=LinkValueStash(link_val_stash_dict), standoff_stash=None)
    upload_state = UploadState(xml_resources.copy(), deepcopy(stash), UploadConfig(), {})
    con = Mock(spec_set=ConnectionLive)
    post_responses = [
        {"@id": "foo_1_iri", "rdfs:label": "foo_1_label"},
        {"@id": "foo_2_iri", "rdfs:label": "foo_2_label"},
        {},  # uploading a stash doesn't rely on a certain response
        {},  # uploading a stash doesn't rely on a certain response
    ]
    con.post = Mock(side_effect=post_responses)
    project_client = ProjectClientStub(con, "1234", None)
    clients = UploadClients(ingest_client_mock, project_client, ListClientMock())

    _upload_resources(clients, upload_state)

    assert len(con.post.call_args_list) == len(post_responses)
    match con.post.call_args_list[2].kwargs:
        case {
            "route": "/v2/values",
            "data": {
                "@type": "my_onto:foo_1_type",
                "@id": "foo_1_iri",
                "@context": dict(),
                "my_onto:hasCustomLinkValue": {
                    "@type": "knora-api:LinkValue",
                    "knora-api:linkValueHasTargetIri": {"@id": "foo_2_iri"},
                },
            },
        }:
            assert True
        case _:
            pytest.fail("POST request was not sent correctly")
    assert not upload_state.pending_resources
    assert not upload_state.failed_uploads
    assert upload_state.iri_resolver.lookup == {"foo_1_id": "foo_1_iri", "foo_2_id": "foo_2_iri"}
    assert not upload_state.pending_stash or upload_state.pending_stash.is_empty()


def test_5_resources_with_stash_and_interrupt_after_2(ingest_client_mock: AssetClient) -> None:
    xml_strings = [
        '<resource label="foo_1_label" restype=":foo_1_type" id="foo_1_id"></resource>',
        '<resource label="foo_2_label" restype=":foo_2_type" id="foo_2_id"></resource>',
        '<resource label="foo_3_label" restype=":foo_3_type" id="foo_3_id"></resource>',
        '<resource label="foo_4_label" restype=":foo_4_type" id="foo_4_id"></resource>',
        '<resource label="foo_5_label" restype=":foo_5_type" id="foo_5_id"></resource>',
    ]
    xml_resources = [XMLResource.from_node(etree.fromstring(xml_str), "my_onto") for xml_str in xml_strings]
    link_val_stash_dict = {
        "foo_1_id": [LinkValueStashItem("foo_1_id", "my_onto:foo_1_type", "my_onto:hasCustomLink", "foo_2_id")],
        "foo_2_id": [LinkValueStashItem("foo_2_id", "my_onto:foo_2_type", "my_onto:hasCustomLink", "foo_1_id")],
    }
    stash = Stash(link_value_stash=LinkValueStash(link_val_stash_dict), standoff_stash=None)
    upload_config = UploadConfig(interrupt_after=2)
    upload_state = UploadState(xml_resources.copy(), deepcopy(stash), upload_config, {})
    con = Mock(spec_set=ConnectionLive)
    post_responses = [
        {"@id": "foo_1_iri", "rdfs:label": "foo_1_label"},
        {"@id": "foo_2_iri", "rdfs:label": "foo_2_label"},
        {"@id": "foo_3_iri", "rdfs:label": "foo_3_label"},
        {"@id": "foo_4_iri", "rdfs:label": "foo_4_label"},
        {"@id": "foo_5_iri", "rdfs:label": "foo_5_label"},
        {},  # uploading a stash doesn't rely on a certain response
        {},  # uploading a stash doesn't rely on a certain response
    ]
    con.post = Mock(side_effect=post_responses)
    project_client = ProjectClientStub(con, "1234", None)
    err_msg = "Interrupted: Maximum number of resources was reached (2)"
    client = UploadClients(ingest_client_mock, project_client, ListClientMock())

    with patch("dsp_tools.commands.xmlupload.xmlupload._handle_upload_error") as _handle_upload_error:
        _upload_resources(client, upload_state)
        iri_resolver_expected = IriResolver({"foo_1_id": "foo_1_iri", "foo_2_id": "foo_2_iri"})
        upload_state_expected = UploadState(xml_resources[2:], stash, upload_config, {}, [], iri_resolver_expected)
        _handle_upload_error.assert_called_once_with(XmlUploadInterruptedError(err_msg), upload_state_expected)

    with patch("dsp_tools.commands.xmlupload.xmlupload._handle_upload_error") as _handle_upload_error:
        _upload_resources(client, upload_state)
        iri_resolver_expected.lookup.update({"foo_3_id": "foo_3_iri", "foo_4_id": "foo_4_iri"})
        upload_state_expected = UploadState(xml_resources[4:], stash, upload_config, {}, [], iri_resolver_expected)
        _handle_upload_error.assert_called_once_with(XmlUploadInterruptedError(err_msg), upload_state_expected)

    with patch("dsp_tools.commands.xmlupload.xmlupload._handle_upload_error") as _handle_upload_error:
        _upload_resources(client, upload_state)
        iri_resolver_expected.lookup.update({"foo_5_id": "foo_5_iri"})
        empty_stash = Stash(standoff_stash=None, link_value_stash=LinkValueStash({}))
        upload_state_expected = UploadState([], empty_stash, upload_config, {}, [], iri_resolver_expected)
        _handle_upload_error.assert_not_called()
        assert upload_state == upload_state_expected


def test_6_resources_with_stash_and_interrupt_after_2(ingest_client_mock: AssetClient) -> None:
    xml_strings = [
        '<resource label="foo_1_label" restype=":foo_1_type" id="foo_1_id"></resource>',
        '<resource label="foo_2_label" restype=":foo_2_type" id="foo_2_id"></resource>',
        '<resource label="foo_3_label" restype=":foo_3_type" id="foo_3_id"></resource>',
        '<resource label="foo_4_label" restype=":foo_4_type" id="foo_4_id"></resource>',
        '<resource label="foo_5_label" restype=":foo_5_type" id="foo_5_id"></resource>',
        '<resource label="foo_6_label" restype=":foo_6_type" id="foo_6_id"></resource>',
    ]
    xml_resources = [XMLResource.from_node(etree.fromstring(xml_str), "my_onto") for xml_str in xml_strings]
    link_val_stash_dict = {
        "foo_1_id": [LinkValueStashItem("foo_1_id", "my_onto:foo_1_type", "my_onto:hasCustomLink", "foo_2_id")],
        "foo_2_id": [LinkValueStashItem("foo_2_id", "my_onto:foo_2_type", "my_onto:hasCustomLink", "foo_1_id")],
    }
    stash = Stash(link_value_stash=LinkValueStash(link_val_stash_dict), standoff_stash=None)
    upload_config = UploadConfig(interrupt_after=2)
    upload_state = UploadState(xml_resources.copy(), deepcopy(stash), upload_config, {})
    con = Mock(spec_set=ConnectionLive)
    post_responses = [
        {"@id": "foo_1_iri", "rdfs:label": "foo_1_label"},
        {"@id": "foo_2_iri", "rdfs:label": "foo_2_label"},
        {"@id": "foo_3_iri", "rdfs:label": "foo_3_label"},
        {"@id": "foo_4_iri", "rdfs:label": "foo_4_label"},
        {"@id": "foo_5_iri", "rdfs:label": "foo_5_label"},
        {"@id": "foo_6_iri", "rdfs:label": "foo_6_label"},
        {},  # uploading a stash doesn't rely on a certain response
        {},  # uploading a stash doesn't rely on a certain response
    ]
    con.post = Mock(side_effect=post_responses)
    project_client = ProjectClientStub(con, "1234", None)
    err_msg = "Interrupted: Maximum number of resources was reached (2)"
    client = UploadClients(ingest_client_mock, project_client, ListClientMock())

    with patch("dsp_tools.commands.xmlupload.xmlupload._handle_upload_error") as _handle_upload_error:
        _upload_resources(client, upload_state)
        iri_resolver_expected = IriResolver({"foo_1_id": "foo_1_iri", "foo_2_id": "foo_2_iri"})
        upload_state_expected = UploadState(xml_resources[2:], stash, upload_config, {}, [], iri_resolver_expected)
        _handle_upload_error.assert_called_once_with(XmlUploadInterruptedError(err_msg), upload_state_expected)

    with patch("dsp_tools.commands.xmlupload.xmlupload._handle_upload_error") as _handle_upload_error:
        _upload_resources(client, upload_state)
        iri_resolver_expected.lookup.update({"foo_3_id": "foo_3_iri", "foo_4_id": "foo_4_iri"})
        upload_state_expected = UploadState(xml_resources[4:], stash, upload_config, {}, [], iri_resolver_expected)
        _handle_upload_error.assert_called_once_with(XmlUploadInterruptedError(err_msg), upload_state_expected)

    with patch("dsp_tools.commands.xmlupload.xmlupload._handle_upload_error") as _handle_upload_error:
        _upload_resources(client, upload_state)
        iri_resolver_expected.lookup.update({"foo_5_id": "foo_5_iri", "foo_6_id": "foo_6_iri"})
        upload_state_expected = UploadState([], stash, upload_config, {}, [], iri_resolver_expected)
        _handle_upload_error.assert_called_once_with(XmlUploadInterruptedError(err_msg), upload_state_expected)

    with patch("dsp_tools.commands.xmlupload.xmlupload._handle_upload_error") as _handle_upload_error:
        _upload_resources(client, upload_state)
        empty_stash = Stash(standoff_stash=None, link_value_stash=LinkValueStash({}))
        upload_state_expected = UploadState([], empty_stash, upload_config, {}, [], iri_resolver_expected)
        _handle_upload_error.assert_not_called()
        assert upload_state == upload_state_expected


def test_logging(caplog: pytest.LogCaptureFixture, ingest_client_mock: AssetClient) -> None:
    xml_strings = [
        '<resource label="foo_1_label" restype=":foo_1_type" id="foo_1_id"></resource>',
        '<resource label="foo_2_label" restype=":foo_2_type" id="foo_2_id"></resource>',
        '<resource label="foo_3_label" restype=":foo_3_type" id="foo_3_id"></resource>',
        '<resource label="foo_4_label" restype=":foo_4_type" id="foo_4_id"></resource>',
        '<resource label="foo_5_label" restype=":foo_5_type" id="foo_5_id"></resource>',
    ]
    xml_resources = [XMLResource.from_node(etree.fromstring(xml_str), "my_onto") for xml_str in xml_strings]
    link_val_stash_dict = {
        "foo_1_id": [LinkValueStashItem("foo_1_id", "my_onto:foo_1_type", "my_onto:hasCustomLink", "foo_2_id")],
        "foo_2_id": [LinkValueStashItem("foo_2_id", "my_onto:foo_2_type", "my_onto:hasCustomLink", "foo_1_id")],
    }
    stash = Stash(link_value_stash=LinkValueStash(link_val_stash_dict), standoff_stash=None)
    upload_config = UploadConfig(interrupt_after=2)
    upload_state = UploadState(xml_resources.copy(), deepcopy(stash), upload_config, {})
    con = Mock(spec_set=ConnectionLive)
    post_responses = [
        {"@id": "foo_1_iri", "rdfs:label": "foo_1_label"},
        {"@id": "foo_2_iri", "rdfs:label": "foo_2_label"},
        {"@id": "foo_3_iri", "rdfs:label": "foo_3_label"},
        {"@id": "foo_4_iri", "rdfs:label": "foo_4_label"},
        {"@id": "foo_5_iri", "rdfs:label": "foo_5_label"},
        {},  # uploading a stash doesn't rely on a certain response
        {},  # uploading a stash doesn't rely on a certain response
    ]
    con.post = Mock(side_effect=post_responses)
    project_client = ProjectClientStub(con, "1234", None)
    clients = UploadClients(ingest_client_mock, project_client, ListClientMock())

    with patch("dsp_tools.commands.xmlupload.xmlupload._handle_upload_error"):
        _upload_resources(clients, upload_state)
        assert caplog.records[1].message == "Created resource 1/5: 'foo_1_label' (ID: 'foo_1_id', IRI: 'foo_1_iri')"
        assert caplog.records[3].message == "Created resource 2/5: 'foo_2_label' (ID: 'foo_2_id', IRI: 'foo_2_iri')"
        caplog.clear()

        _upload_resources(clients, upload_state)
        assert caplog.records[1].message == "Created resource 3/5: 'foo_3_label' (ID: 'foo_3_id', IRI: 'foo_3_iri')"
        assert caplog.records[3].message == "Created resource 4/5: 'foo_4_label' (ID: 'foo_4_id', IRI: 'foo_4_iri')"
        caplog.clear()

        _upload_resources(clients, upload_state)
        assert caplog.records[1].message == "Created resource 5/5: 'foo_5_label' (ID: 'foo_5_id', IRI: 'foo_5_iri')"
        assert caplog.records[3].message == "  Upload resptrs of resource 'foo_1_id'..."
        assert caplog.records[5].message == "  Upload resptrs of resource 'foo_2_id'..."
        caplog.clear()


def test_post_requests(ingest_client_mock: AssetClient) -> None:
    xml_strings = [
        '<resource label="foo_1_label" restype=":foo_1_type" id="foo_1_id"></resource>',
        '<resource label="foo_2_label" restype=":foo_2_type" id="foo_2_id"></resource>',
        '<resource label="foo_3_label" restype=":foo_3_type" id="foo_3_id"></resource>',
        '<resource label="foo_4_label" restype=":foo_4_type" id="foo_4_id"></resource>',
        '<resource label="foo_5_label" restype=":foo_5_type" id="foo_5_id"></resource>',
        '<resource label="foo_6_label" restype=":foo_6_type" id="foo_6_id"></resource>',
    ]
    xml_resources = [XMLResource.from_node(etree.fromstring(xml_str), "my_onto") for xml_str in xml_strings]
    link_val_stash_dict = {
        "foo_1_id": [LinkValueStashItem("foo_1_id", "my_onto:foo_1_type", "my_onto:hasCustomLink", "foo_2_id")],
        "foo_2_id": [LinkValueStashItem("foo_2_id", "my_onto:foo_2_type", "my_onto:hasCustomLink", "foo_1_id")],
    }
    stash = Stash(link_value_stash=LinkValueStash(link_val_stash_dict), standoff_stash=None)
    upload_config = UploadConfig(interrupt_after=2)
    upload_state = UploadState(xml_resources.copy(), deepcopy(stash), upload_config, {})
    con = Mock(spec_set=ConnectionLive)
    post_responses = [
        {"@id": "foo_1_iri", "rdfs:label": "foo_1_label"},
        {"@id": "foo_2_iri", "rdfs:label": "foo_2_label"},
        {"@id": "foo_3_iri", "rdfs:label": "foo_3_label"},
        {"@id": "foo_4_iri", "rdfs:label": "foo_4_label"},
        {"@id": "foo_5_iri", "rdfs:label": "foo_5_label"},
        {"@id": "foo_6_iri", "rdfs:label": "foo_6_label"},
        {},  # uploading a stash doesn't rely on a certain response
        {},  # uploading a stash doesn't rely on a certain response
    ]
    con.post = Mock(side_effect=post_responses)
    project_client = ProjectClientStub(con, "1234", None)
    clients = UploadClients(ingest_client_mock, project_client, ListClientMock())

    with patch("dsp_tools.commands.xmlupload.xmlupload._handle_upload_error"):
        _upload_resources(clients, upload_state)
        _upload_resources(clients, upload_state)
        _upload_resources(clients, upload_state)
        _upload_resources(clients, upload_state)
        assert len(con.post.call_args_list) == len(post_responses)


def test_interruption_if_resource_cannot_be_created_because_of_404() -> None:
    xml_strings = [
        '<resource label="foo_1_label" restype=":foo_1_type" id="foo_1_id"></resource>',
        '<resource label="foo_2_label" restype=":foo_2_type" id="foo_2_id"></resource>',
    ]
    xml_resources = [XMLResource.from_node(etree.fromstring(xml_str), "my_onto") for xml_str in xml_strings]
    upload_state = UploadState(xml_resources.copy(), Stash(None, None), UploadConfig(), {}, [], IriResolver())
    con = ConnectionLive("foo")
    con._log_request = Mock()  # type: ignore[method-assign]
    con._log_response = Mock()  # type: ignore[method-assign]
    resp_404 = Response()
    resp_404.status_code = 404
    post_responses = [resp_404]
    con.session.request = Mock(side_effect=post_responses)  # type: ignore[method-assign]
    project_client = ProjectClientStub(con, "1234", None)
    ingest_client = DspIngestClientLive("", AuthenticationClientMockBase(), "1234", ".")

    with patch("dsp_tools.commands.xmlupload.xmlupload._handle_upload_error") as _handle_upload_error:
        _upload_resources(UploadClients(ingest_client, project_client, ListClientMock()), upload_state)
        msg = (
            "Lost connection to DSP server, probably because the server is down. "
            "Please continue later with 'resume-xmlupload'. Reason for this failure: "
            "Permanently unable to execute the network action. "
        )
        assert len(_handle_upload_error.call_args_list) == 1
        err_actual: XmlUploadInterruptedError = _handle_upload_error.call_args_list[0].args[0]
        upload_state_actual: UploadState = _handle_upload_error.call_args_list[0].args[1]
        assert msg in err_actual.message
        assert upload_state_actual == upload_state



from typing import Callable
from typing import Sequence
from typing import cast

from lxml import etree

from dsp_tools.commands.validate_data.models.data_deserialised import AbstractFileValueDeserialised
from dsp_tools.commands.validate_data.models.data_deserialised import BitstreamDeserialised
from dsp_tools.commands.validate_data.models.data_deserialised import BooleanValueDeserialised
from dsp_tools.commands.validate_data.models.data_deserialised import ColorValueDeserialised
from dsp_tools.commands.validate_data.models.data_deserialised import DataDeserialised
from dsp_tools.commands.validate_data.models.data_deserialised import DateValueDeserialised
from dsp_tools.commands.validate_data.models.data_deserialised import DecimalValueDeserialised
from dsp_tools.commands.validate_data.models.data_deserialised import GeonameValueDeserialised
from dsp_tools.commands.validate_data.models.data_deserialised import IIIFUriDeserialised
from dsp_tools.commands.validate_data.models.data_deserialised import IntValueDeserialised
from dsp_tools.commands.validate_data.models.data_deserialised import LinkValueDeserialised
from dsp_tools.commands.validate_data.models.data_deserialised import ListValueDeserialised
from dsp_tools.commands.validate_data.models.data_deserialised import ProjectDeserialised
from dsp_tools.commands.validate_data.models.data_deserialised import ProjectInformation
from dsp_tools.commands.validate_data.models.data_deserialised import ResourceDeserialised
from dsp_tools.commands.validate_data.models.data_deserialised import RichtextDeserialised
from dsp_tools.commands.validate_data.models.data_deserialised import SimpleTextDeserialised
from dsp_tools.commands.validate_data.models.data_deserialised import TimeValueDeserialised
from dsp_tools.commands.validate_data.models.data_deserialised import UriValueDeserialised
from dsp_tools.commands.validate_data.models.data_deserialised import ValueDeserialised

KNORA_API = "http://api.knora.org/ontology/knora-api/v2#"
REGION_RESOURCE = KNORA_API + "Region"
LINKOBJ_RESOURCE = KNORA_API + "LinkObj"
VIDEO_SEGMENT_RESOURCE = KNORA_API + "VideoSegment"
AUDIO_SEGMENT_RESOURCE = KNORA_API + "AudioSegment"


def deserialise_xml(root: etree._Element) -> ProjectDeserialised:
    """
    Takes the root of an XML
    Extracts the data of the project and transforms all its resources.

    Args:
        root: root of an xml with qnames and comments removed

    Returns:
        Class instance with the information reformatted
    """
    shortcode = root.attrib["shortcode"]
    default_ontology = root.attrib["default-ontology"]
    project_info = ProjectInformation(shortcode, default_ontology)
    data_deserialised = _deserialise_all_resources(root)
    return ProjectDeserialised(project_info, data_deserialised)


def _deserialise_all_resources(root: etree._Element) -> DataDeserialised:
    all_res: list[ResourceDeserialised] = []
    for res in root.iterdescendants(tag="resource"):
        res_id = res.attrib["id"]
        lbl = cast(str, res.attrib.get("label"))
        dsp_type = None
        res_type = res.attrib["restype"]
        if res_type == REGION_RESOURCE:
            dsp_type = REGION_RESOURCE
        elif res_type == VIDEO_SEGMENT_RESOURCE:
            dsp_type = VIDEO_SEGMENT_RESOURCE
        elif res_type == AUDIO_SEGMENT_RESOURCE:
            dsp_type = AUDIO_SEGMENT_RESOURCE
        if dsp_type:
            all_res.append(ResourceDeserialised(res_id, lbl, dsp_type, []))
        else:
            all_res.append(_deserialise_one_resource(res))
    file_values = _deserialise_file_value(root)
    return DataDeserialised(all_res, file_values)


def _deserialise_one_resource(resource: etree._Element) -> ResourceDeserialised:
    values: list[ValueDeserialised] = []
    for val in resource.iterchildren():
        values.extend(_deserialise_one_property(val))
    return ResourceDeserialised(
        res_id=resource.attrib["id"],
        res_class=resource.attrib["restype"],
        label=resource.attrib["label"],
        values=values,
    )


def _deserialise_one_property(prop_ele: etree._Element) -> Sequence[ValueDeserialised]:  # noqa: PLR0911 (too-many-branches, return statements)
    match prop_ele.tag:
        case "boolean-prop":
            return _deserialise_one_value(prop_ele, BooleanValueDeserialised)
        case "color-prop":
            return _deserialise_one_value(prop_ele, ColorValueDeserialised)
        case "date-prop":
            return _deserialise_one_value(prop_ele, DateValueDeserialised)
        case "decimal-prop":
            return _deserialise_one_value(prop_ele, DecimalValueDeserialised)
        case "geoname-prop":
            return _deserialise_one_value(prop_ele, GeonameValueDeserialised)
        case "list-prop":
            return _deserialise_list_prop(prop_ele)
        case "integer-prop":
            return _deserialise_one_value(prop_ele, IntValueDeserialised)
        case "resptr-prop":
            return _deserialise_one_value(prop_ele, LinkValueDeserialised)
        case "text-prop":
            return _deserialise_text_prop(prop_ele)
        case "time-prop":
            return _deserialise_one_value(prop_ele, TimeValueDeserialised)
        case "uri-prop":
            return _deserialise_one_value(prop_ele, UriValueDeserialised)
        case _:
            return []


def _deserialise_one_value(
    prop: etree._Element, func: Callable[[str, str | None], ValueDeserialised]
) -> Sequence[ValueDeserialised]:
    prop_name = prop.attrib["name"]
    return [func(prop_name, x.text) for x in prop.iterchildren()]


def _deserialise_list_prop(prop: etree._Element) -> list[ListValueDeserialised]:
    prop_name = prop.attrib["name"]
    list_name = prop.attrib["list"]
    return [ListValueDeserialised(prop_name, x.text, list_name) for x in prop.iterchildren()]


def _deserialise_text_prop(prop: etree._Element) -> list[SimpleTextDeserialised | RichtextDeserialised]:
    prop_name = prop.attrib["name"]
    all_vals: list[SimpleTextDeserialised | RichtextDeserialised] = []
    for val in prop.iterchildren():
        match val.attrib["encoding"]:
            case "utf8":
                all_vals.append(SimpleTextDeserialised(prop_name, _get_text_as_string(val)))
            case "xml":
                all_vals.append(RichtextDeserialised(prop_name, _get_text_as_string(val)))
    return all_vals


def _get_text_as_string(value: etree._Element) -> str | None:
    if len(value):
        text_list = []
        if found := value.text:
            text_list = [found]
        text_list.extend([etree.tostring(child, encoding="unicode", method="xml") for child in value])
        return "".join(text_list).strip()
    else:
        return value.text


def _deserialise_file_value(root: etree._Element) -> list[AbstractFileValueDeserialised]:
    file_values: list[AbstractFileValueDeserialised] = []
    for res in root.iterchildren(tag="resource"):
        res_id = res.attrib["id"]
        if (bitstream := res.find("bitstream")) is not None:
            file_values.append(BitstreamDeserialised(res_id, bitstream.text))
        elif (iiif_uri := res.find("iiif-uri")) is not None:
            file_values.append(IIIFUriDeserialised(res_id, iiif_uri.text))
    return file_values



import pytest

from dsp_tools.commands.project.models.project import Project
from dsp_tools.models.langstring import LangString
from dsp_tools.models.langstring import Languages
from test.integration.commands.xmlupload.connection_mock import ConnectionMockBase


@pytest.fixture
def project() -> Project:
    return Project(
        con=ConnectionMockBase(),
        iri="http://rdfh.ch/test",
        shortcode="0FF0",
        shortname="test_project",
        longname="Test Project",
        description=LangString({Languages.EN: "This is a test project", Languages.DE: "Das ist ein Testprojekt"}),
        keywords=set(),
        selfjoin=False,
        status=True,
        logo="logo.gif",
    )


def test_return_values(project: Project) -> None:
    assert project.iri == "http://rdfh.ch/test"
    assert project.shortcode == "0FF0"
    assert project.shortname == "test_project"
    assert project.longname == "Test Project"
    assert project.description["en"] == "This is a test project"
    assert project.description["de"] == "Das ist ein Testprojekt"
    assert project.selfjoin is False
    assert project.status is True
    assert project.keywords == set()


def test_toJsonObj_create(project: Project) -> None:
    res_json = project._toJsonObj_create()
    expected = {
        "shortcode": "0FF0",
        "shortname": "test_project",
        "longname": "Test Project",
        "description": [
            {"language": "en", "value": "This is a test project"},
            {"language": "de", "value": "Das ist ein Testprojekt"},
        ],
        "selfjoin": False,
        "status": True,
    }
    assert res_json == expected


if __name__ == "__main__":
    pytest.main([__file__])



from __future__ import annotations

import copy
from copy import deepcopy
from pathlib import Path

from loguru import logger
from lxml import etree

from dsp_tools.models.exceptions import InputError


def parse_and_clean_xml_file(input_file: Path) -> etree._Element:
    """
    Parse an XML file with DSP-conform data,
    remove namespace URI from the elements' names,
    and transform the special tags `<region>`, `<link>`, `<video-segment>`, `<audio-segment>`
    to their technically correct form
    `<resource restype="Region">`, `<resource restype="LinkObj">`,
    `<resource restype="VideoSegment">`, `<resource restype="AudioSegment">`.

    Args:
        input_file: path to the XML file

    Returns:
        the root element of the parsed XML file

    Raises:
        InputError: if the input is not of either the expected types
    """
    root = parse_xml_file(input_file)
    root = remove_comments_from_element_tree(root)
    return transform_special_tags_make_localname(root)


def transform_special_tags_make_localname(input_tree: etree._Element) -> etree._Element:
    """
    This function removes the namespace URIs from the elements' names
    and transforms the special tags `<region>`, `<link>`, `<video-segment>`, `<audio-segment>`
    to their technically correct form
    `<resource restype="Region">`, `<resource restype="LinkObj">`,
    `<resource restype="VideoSegment">`, `<resource restype="AudioSegment">`.

    Args:
        input_tree: unclean tree

    Returns:
        cleaned tree
    """
    tree = transform_into_localnames(input_tree)
    return _transform_special_tags(tree)


def _transform_special_tags(tree: etree._Element) -> etree._Element:
    tree = deepcopy(tree)
    for elem in tree.iter():
        if elem.tag == "link":
            elem.attrib["restype"] = "LinkObj"
            elem.tag = "resource"
        elif elem.tag == "region":
            elem.attrib["restype"] = "Region"
            elem.tag = "resource"
        elif elem.tag == "video-segment":
            elem.attrib["restype"] = "VideoSegment"
            _correct_is_segment_of_property(elem, "VideoSegment")
            elem.tag = "resource"
        elif elem.tag == "audio-segment":
            elem.attrib["restype"] = "AudioSegment"
            _correct_is_segment_of_property(elem, "AudioSegment")
            elem.tag = "resource"
    return tree


def _correct_is_segment_of_property(segment: etree._Element, restype: str) -> None:
    for child in segment.iterchildren():
        if child.tag == "isSegmentOf":
            child.tag = f"is{restype}Of"
            break


def transform_into_localnames(root: etree._Element) -> etree._Element:
    """
    This function removes the namespace URIs from the elements' names

    Args:
        root: unclean tree

    Returns:
        cleaned tree
    """
    tree = deepcopy(root)
    for elem in tree.iter():
        elem.tag = etree.QName(elem).localname
    return tree


def remove_comments_from_element_tree(input_tree: etree._Element) -> etree._Element:
    """
    This function removes comments and processing instructions.
    Commented out properties break the XMLProperty constructor.

    Args:
        input_tree: etree root that will be cleaned

    Returns:
        clean xml
    """
    root = copy.deepcopy(input_tree)
    for c in root.xpath("//comment()"):
        c.getparent().remove(c)
    for c in root.xpath("//processing-instruction()"):
        c.getparent().remove(c)
    return root


def parse_xml_file(input_file: str | Path) -> etree._Element:
    """
    This function parses an XML file and returns an Element Tree

    Args:
        input_file: path to the input file

    Returns:
        element tree

    Raises:
        InputError: if the file contains a syntax error
    """
    parser = etree.XMLParser(remove_comments=True, remove_pis=True)
    try:
        return etree.parse(source=input_file, parser=parser).getroot()
    except etree.XMLSyntaxError as err:
        logger.opt(exception=True).error(f"The XML file contains the following syntax error: {err.msg}")
        raise InputError(f"The XML file contains the following syntax error: {err.msg}") from None


def remove_namespaces_from_xml(data_xml: etree._Element) -> etree._Element:
    """
    This function removes all the namespaces from an XML element tree.

    Args:
        data_xml: xml with namespaces

    Returns:
        the XMl without the namespaces
    """
    xml_no_namespace = copy.deepcopy(data_xml)
    for elem in xml_no_namespace.iter():
        if not isinstance(elem, (etree._Comment, etree._ProcessingInstruction)):
            elem.tag = etree.QName(elem).localname
    return xml_no_namespace



from dsp_tools.xmllib.models.dsp_base_resources import RegionResource
from dsp_tools.xmllib.models.resource import Resource
from dsp_tools.xmllib.models.root import XMLRoot


def test_root_add_resources() -> None:
    xml_root = XMLRoot.create_new("0000", "test")

    # add_resource
    region_res = RegionResource.create_new("regionID", "label", "id1")
    region_res.add_rectangle((0.1, 0.1), (0.2, 0.2))
    xml_root.add_resource(region_res)
    assert len(xml_root.resources) == 1

    # add_resource_multiple
    many_resources = [Resource.create_new("id1", ":Restype", "label"), Resource.create_new("id2", ":Restype", "label")]
    xml_root.add_resource_multiple(many_resources)
    assert len(xml_root.resources) == 3

    xml_root.add_resource_multiple([])
    assert len(xml_root.resources) == 3

    # add_resource_optional
    no_resource = None
    xml_root.add_resource_optional(no_resource)
    assert len(xml_root.resources) == 3

    serialised = xml_root.serialise()
    region = list(serialised.iterdescendants(tag="{https://dasch.swiss/schema}region"))
    assert len(region) == 1
    general_resources = list(serialised.iterdescendants(tag="{https://dasch.swiss/schema}resource"))
    assert len(general_resources) == 2



from typing import Any

from dsp_tools.models.exceptions import InputError
from dsp_tools.xmllib.models.config_options import NewlineReplacement


def convert_to_bool(value: Any) -> bool:
    """
    Turns a value into a bool string, suitable for an XML.
    It is case-insensitive, meaning that the words can also be capitalised.

    Accepted values:
         - `false`, `0`, `0.0`, `no`, `non`, `nein` -> `False`
         - `true`, `1`, `1.0`, `yes`, `oui`, `ja` -> `True`

    Args:
        value: value to transform

    Returns:
        `True` or `False` if it is an accepted value.

    Raises:
        InputError: If the value is not convertable to a boolean

    Examples:
        ```python
        result = xmllib.convert_to_bool_string(1)
        # result == True
        ```

        ```python
        result = xmllib.convert_to_bool_string("nein")
        # result == False
        ```

        ```python
        # because this is not an accepted value, it is returned as a string

        result = xmllib.convert_to_bool_string(None)
        # raises InputError
        ```
    """
    str_val = str(value).lower().strip()
    if str_val in ("false", "0", "0.0", "no", "non", "nein"):
        return False
    elif str_val in ("true", "1", "1.0", "yes", "oui", "ja"):
        return True
    raise InputError(f"The entered value '{value}' cannot be converted to a bool.")


def replace_newlines_with_tags(text: str, converter_option: NewlineReplacement) -> str:
    """
    Converts the newlines in a string to XML tags.

    Args:
        text: string to convert
        converter_option: specifies what tag to use instead of the newline

    Returns:
        String with replaced values

    Raises:
        InputError: If an invalid conversion option is given

    Examples:
        ```python
        result = xmllib.replace_newlines_with_tags(
            "Start\\nEnd", xmllib.NewlineReplacement.NONE
        )
        # result == "Start\\nEnd"
        ```

        ```python
        result = xmllib.replace_newlines_with_tags(
            "Start\\nEnd", xmllib.NewlineReplacement.LINEBREAK
        )
        # result == "Start<br/>End"
        ```

        ```python
        result = xmllib.replace_newlines_with_tags(
            "Start\\n\\nEnd", xmllib.NewlineReplacement.PARAGRAPH
        )
        # result == "<p>Start</p><p>End</p>"
        ```
    """
    match converter_option:
        case NewlineReplacement.NONE:
            return text
        case NewlineReplacement.LINEBREAK:
            return replace_newlines_with_br_tags(text)
        case NewlineReplacement.PARAGRAPH:
            return replace_newlines_with_paragraph_tags(text)


def replace_newlines_with_paragraph_tags(text: str) -> str:
    """
    Replace `Start\\nEnd` with `<p>Start</p><p>End</p>`

    Args:
        text: string to be formatted

    Returns:
        Formatted string with paragraph tags

    Examples:
        ```python
        result = xmllib.replace_newlines_with_paragraph_tags("Start\\nEnd")
        # result == "<p>Start</p><p>End</p>"
        ```

        ```python
        # multiple consecutive newlines will be treated as one newline

        result = xmllib.replace_newlines_with_paragraph_tags("Start\\n\\nEnd")
        # result == "<p>Start</p><p>End</p>"
        ```
    """
    splt = [x for x in text.split("\n") if x != ""]
    formatted = [f"<p>{x}</p>" for x in splt]
    return "".join(formatted)


def replace_newlines_with_br_tags(text: str) -> str:
    """
    Replaces `\\n` with `<br/>`

    Args:
        text: string to be formatted

    Returns:
        Formatted string with break-line tags

    Examples:
        ```python
        result = xmllib.replace_newlines_with_br_tags("Start\\nEnd")
        # result == "Start<br/>End"
        ```

        ```python
        # multiple consecutive newlines will be converted into multiple break-lines

        result = xmllib.replace_newlines_with_br_tags("Start\\n\\nEnd")
        # result == "Start<br/><br/>End"
        ```
    """
    return text.replace("\n", "<br/>")



from typing import Any
from unittest.mock import Mock
from unittest.mock import patch

import pytest

from dsp_tools.commands.validate_data.api_clients import ListClient
from dsp_tools.commands.validate_data.api_clients import OntologyClient
from dsp_tools.commands.validate_data.api_connection import ApiConnection
from dsp_tools.models.exceptions import InternalError
from dsp_tools.models.exceptions import UserError


@pytest.fixture
def api_con() -> ApiConnection:
    return ApiConnection("http://0.0.0.0:3333")


@pytest.fixture
def ontology_connection(api_con: ApiConnection) -> OntologyClient:
    return OntologyClient(api_con, "9999")


@pytest.fixture
def list_connection(api_con: ApiConnection) -> ListClient:
    return ListClient(api_con, "9999")


class TestOntologyConnection:
    def test_get_ontology_iris_ok(self, ontology_connection: OntologyClient) -> None:
        mock_response = Mock()
        mock_response.ok = True
        mock_response.json.return_value = {"project": {"ontologies": ["onto_iri"]}}
        with patch.object(ontology_connection.api_con, "get_with_endpoint", return_value=mock_response) as patched_get:
            result = ontology_connection._get_ontology_iris()
            assert result == ["onto_iri"]
            patched_get.assert_called_once_with(endpoint="admin/projects/shortcode/9999")

    def test_get_ontology_iris_non_ok_code(self, ontology_connection: OntologyClient) -> None:
        mock_response = Mock()
        mock_response.ok = False
        with patch.object(ontology_connection.api_con, "get_with_endpoint", return_value=mock_response) as patched_get:
            with pytest.raises(InternalError):
                ontology_connection._get_ontology_iris()
            patched_get.assert_called_once_with(endpoint="admin/projects/shortcode/9999")

    def test_get_ontology_iris_no_ontology_key(self, ontology_connection: OntologyClient) -> None:
        mock_response = Mock()
        mock_response.ok = True
        mock_response.json.return_value = {}
        with patch.object(ontology_connection.api_con, "get_with_endpoint", return_value=mock_response) as patched_get:
            with pytest.raises(UserError):
                ontology_connection._get_ontology_iris()
            patched_get.assert_called_once_with(endpoint="admin/projects/shortcode/9999")

    def test_get_one_ontology(self, ontology_connection: OntologyClient) -> None:
        mock_response = Mock()
        mock_response.ok = True
        mock_response.text = "Turtle Text"
        with patch.object(ontology_connection.api_con, "get_with_url", return_value=mock_response) as patched_get:
            result = ontology_connection._get_one_ontology("iri")
            assert result == "Turtle Text"
            patched_get.assert_called_once_with(url="iri", headers={"Accept": "text/turtle"})


class TestListConnection:
    def test_get_all_list_iris(self, list_connection: ListClient) -> None:
        mock_response = Mock()
        mock_response.ok = True
        mock_response.json.return_value = {"lists": []}
        with patch.object(list_connection.api_con, "get_with_endpoint", return_value=mock_response) as patched_get:
            result = list_connection._get_all_list_iris()
            assert result == {"lists": []}
            patched_get.assert_called_once_with(endpoint="admin/lists?9999")

    def test_get_all_list_iris_non_ok_code(self, list_connection: ListClient) -> None:
        mock_response = Mock()
        mock_response.ok = False
        with patch.object(list_connection.api_con, "get_with_endpoint", return_value=mock_response) as patched_get:
            with pytest.raises(InternalError):
                list_connection._get_all_list_iris()
            patched_get.assert_called_once_with(endpoint="admin/lists?9999")

    def test_get_one_list(self, list_connection: ListClient) -> None:
        mock_response = Mock()
        mock_response.ok = True
        mock_response.json.return_value = {"type": "ListGetResponseADM", "list": {}}
        with patch.object(list_connection.api_con, "get_with_endpoint", return_value=mock_response) as patched_get:
            result = list_connection._get_one_list("http://rdfh.ch/lists/9999/WWqeCEj8R_qrK5djsVcHvg")
            assert result == {"type": "ListGetResponseADM", "list": {}}
            patched_get.assert_called_once_with(
                endpoint="admin/lists/http%3A%2F%2Frdfh.ch%2Flists%2F9999%2FWWqeCEj8R_qrK5djsVcHvg"
            )

    def test_get_one_list_non_ok_code(self, list_connection: ListClient) -> None:
        mock_response = Mock()
        mock_response.ok = False
        with patch.object(list_connection.api_con, "get_with_endpoint", return_value=mock_response) as patched_get:
            with pytest.raises(InternalError):
                list_connection._get_one_list("http://rdfh.ch/lists/9999/WWqeCEj8R_qrK5djsVcHvg")
            patched_get.assert_called_once_with(
                endpoint="admin/lists/http%3A%2F%2Frdfh.ch%2Flists%2F9999%2FWWqeCEj8R_qrK5djsVcHvg"
            )

    def test_extract_list_iris(
        self, list_connection: ListClient, response_all_list_one_project: dict[str, Any]
    ) -> None:
        extracted = list_connection._extract_list_iris(response_all_list_one_project)
        expected = {"http://rdfh.ch/lists/9999/list1", "http://rdfh.ch/lists/9999/list2"}
        assert set(extracted) == expected

    def test_extract_list_iris_no_lists(
        self, list_connection: ListClient, response_all_list_one_project_no_lists: dict[str, Any]
    ) -> None:
        extracted = list_connection._extract_list_iris(response_all_list_one_project_no_lists)
        assert not extracted

    def test_reformat_one_list(self, list_connection: ListClient, response_one_list: dict[str, Any]) -> None:
        reformatted = list_connection._reformat_one_list(response_one_list)
        expected_nodes = {"n1", "n1.1", "n1.1.1", "n1.1.2"}
        assert reformatted.list_iri == "http://rdfh.ch/lists/9999/list1"
        assert reformatted.list_name == "firstList"
        assert set(reformatted.nodes) == expected_nodes


if __name__ == "__main__":
    pytest.main([__file__])
